from copy import deepcopy

import sys

import napari
from PyQt5.QtWidgets import QTextEdit
from magicclass import magicclass, field, bind_key
from magicclass.widgets import FreeWidget
from magicgui.types import Undefined
from magicgui.widgets import TextEdit
from napari.utils.notifications import notification_manager
from qtpy.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QAction, QPlainTextEdit
from qtpy.QtCore import QRegularExpression, Qt, Signal, QObject, QThread
from qtpy.QtGui import QTextCharFormat, QFont, QSyntaxHighlighter, QColor
from qtpy import QtCore, QtGui
import numpy as np
import skimage
import warnings
import keyword
import psygnal
from . import ProgressWidget


def execute_script(script, other_locals=None):
    if other_locals is None:
        other_locals = dict()
    globals_ = {"np": np, "skimage": skimage} | other_locals
    exec(script, globals_, globals_)
    return globals_


class ScriptWorker(QObject):
    done = Signal("PyQt_PyObject")
    script = None
    variables = None
    _output_variables = {}

    def run(self):
        if self.script is None:
            raise ValueError("No script was set!")
        try:
            self._output_variables = execute_script(self.script, self.variables)
            self.done.emit(self._output_variables)
        except Exception as e:
            self.done.emit({"exception": e})

    @property
    def output_variables(self):
        return self._output_variables


class PythonHighlighter(QSyntaxHighlighter):
    class HighlightingRule:
        pattern = QRegularExpression()
        format = QTextCharFormat()

    highlightingRules = []
    commentFormat = QTextCharFormat()
    keywordFormat = QTextCharFormat()
    classFormat = QTextCharFormat()
    quotationFormat = QTextCharFormat()
    singleLineCommentFormat = QTextCharFormat()
    tripleQuotationFormat = QTextCharFormat()
    functionFormat = QTextCharFormat()
    numericFormat = QTextCharFormat()

    def __init__(self, parent):
        super().__init__(parent)
        self.keywordFormat.setForeground(QColor("darkorange"))
        self.keywordFormat.setFontWeight(QFont.Bold)
        keywordPatterns = ["\\b%s\\b" % kw for kw in keyword.kwlist]
        for pattern in keywordPatterns:
            rule = self.HighlightingRule()
            rule.pattern = QRegularExpression(pattern)
            rule.format = self.keywordFormat
            self.highlightingRules.append(rule)

        rule = self.HighlightingRule()
        self.numericFormat.setForeground(QColor("deepskyblue"))
        rule.pattern = QRegularExpression("(0((b[01]+)|(O[0-7]+)|(x[0-9A-Fa-f]+)))|([0-9]*\\.?[0-9]*j?)")
        rule.format = self.numericFormat
        self.highlightingRules.append(rule)

        rule = self.HighlightingRule()
        self.functionFormat.setForeground(QColor("gold"))
        rule.pattern = QRegularExpression("\\b[a-z][A-Za-z0-9_]*(?=\\()")
        rule.format = self.functionFormat
        self.highlightingRules.append(rule)

        rule = self.HighlightingRule()
        self.classFormat.setForeground(QColor("lightblue"))
        rule.pattern = QRegularExpression("\\b[A-Z][A-Za-z0-9_]*(?=\\()")
        rule.format = self.classFormat
        self.highlightingRules.append(rule)

        rule = self.HighlightingRule()
        self.quotationFormat.setForeground(Qt.darkGreen)
        rule.pattern = QRegularExpression("([\"'])(?:(?=(\\\\?))\\2.)*?\\1")
        rule.format = self.quotationFormat
        self.highlightingRules.append(rule)

        rule = self.HighlightingRule()
        self.commentFormat.setForeground(Qt.gray)
        rule.pattern = QRegularExpression("#.*")
        rule.format = self.commentFormat
        self.highlightingRules.append(rule)

    def highlightBlock(self, text):
        for rule in self.highlightingRules:
            matchIterator = rule.pattern.globalMatch(text)
            while matchIterator.hasNext():
                match = matchIterator.next()
                self.setFormat(match.capturedStart(), match.capturedLength(), rule.format)


class CodeEditor(QPlainTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setStyleSheet("font-family:'Courier New'; background-color: black")
        self.highlighter = PythonHighlighter(self.document())

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        if event.key() == QtCore.Qt.Key_Tab:
            self.insertPlainText("    ")
        else:
            super().keyPressEvent(event)


class CodeTextEdit(TextEdit):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.native.setStyleSheet("font-family:'Courier New'; background-color: black")

        def keyPressEvent(*args, **kwargs):
            print(*args, **kwargs)
            # if event.key() == QtCore.Qt.Key_Tab:
            #     self.insertPlainText("    ")
            # else:
            #     super(QTextEdit, self).keyPressEvent(event)

        # self.native.keyPressEvent = keyPressEvent
        self.highlighter = PythonHighlighter(self.native.document())


class ScriptExecuteWidget(FreeWidget):
    changed = psygnal.Signal(dict)

    def __init__(self, editor_key="_script_execute_widget", use_run_button=True, **kwargs):
        super().__init__()
        self.wdt = QWidget()
        layout = QVBoxLayout()
        self.code_editor = CodeEditor()
        layout.addWidget(self.code_editor)
        self.run_button = QPushButton("Run")
        self.run_button.clicked.connect(self.Run)
        layout.addWidget(self.run_button)
        self.wdt.setLayout(layout)
        self.set_widget(self.wdt)
        self.progress_dialog = ProgressWidget(message="Calculating feature, please wait...")
        self._editor_key = editor_key
        self._init_worker()
        self.text_settings = QtCore.QSettings("BIOMAG", "Annotation Toolbox")
        self.code_editor.document().setPlainText(self.text_settings.value(self._editor_key, ""))
        font_size = self.text_settings.value("script_font_size")
        if font_size:
            font = self.code_editor.font()
            font.setPointSize(font_size)
            self.code_editor.setFont(font)
        self._variables = dict()
        self.run_button.setVisible(use_run_button)

    def _init_worker(self):
        self.script_thread = QThread()
        self.script_worker = ScriptWorker()
        self.script_worker.moveToThread(self.script_thread)
        self.script_worker.done.connect(self.script_thread.quit, Qt.DirectConnection)
        self.script_worker.done.connect(lambda: self.run_button.setEnabled(True))
        self.script_worker.done.connect(self.changed.emit, Qt.DirectConnection)
        self.script_thread.started.connect(lambda: self.progress_dialog.setVisible(True))
        self.script_thread.started.connect(self.script_worker.run, Qt.DirectConnection)
        self.script_thread.finished.connect(lambda: self.progress_dialog.setVisible(False))

    @bind_key("Ctrl+Enter")
    def Run(self):
        self.script_thread.wait()
        variables = self.variables
        variables = deepcopy(variables)
        script = self.code_editor.document().toPlainText()
        self.text_settings.setValue(self._editor_key, script)
        self.run_button.setEnabled(False)
        self.script_worker.script = script
        self.script_worker.variables = variables
        self.script_thread.start()

    @property
    def variables(self):
        return self._variables

    # @property
    # def value(self):
    #     return self._variables

    @bind_key("Ctrl-+")
    def _increase_font_size(self):
        font = self.code_editor.font()
        font.setPointSize(font.pointSize() + 1)
        self.code_editor.setFont(font)
        self.text_settings.setValue("script_font_size", font.pointSize())

    @bind_key("Ctrl--")
    def _decrease_font_size(self):
        font = self.code_editor.font()
        font.setPointSize(font.pointSize() - 1)
        self.code_editor.setFont(font)
        self.text_settings.setValue("script_font_size", font.pointSize())



import os
import sys
import tempfile
import atexit
import warnings
import traceback

import numpy as np
import random
import string
import shutil
from .feature_extractor import FeatureExtractor
from .._helper_functions import layer_dims_displayed, layer_dims_not_displayed, layer_ndisplay, layer_slice_indices
from napari import Viewer, layers
from typing import Union, Optional
import glob
TEMP_SUFFIX = "_nd_annotator"


class FeatureManager:
    def __init__(self, viewer):
        self.layer = None
        self.dims_displayed = None
        self.memmaps: list[Optional[Union[np.ndarray, np.memmap]]] = [None, None]
        self.slices_calculated = dict()
        self.clean_tmp()
        self.temp_folder = tempfile.mkdtemp(suffix=TEMP_SUFFIX)
        # map layers to file prefix
        self.prefix_map = dict()
        self.viewer: Viewer = viewer
        self.feature_extractor = FeatureExtractor()
        atexit.register(self.clean)

    def get_features(self, layer: layers.Layer, block=True):
        if layer_ndisplay(layer) == 3:
            return None, None
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            dims_displayed = tuple(layer_dims_displayed(layer))
            dims_not_displayed = tuple(layer_dims_not_displayed(layer))
        if layer != self.layer or dims_displayed != self.dims_displayed:
            self.init_file(layer, dims_displayed)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            slice_indices = layer_slice_indices(layer)
            idx = tuple(slice_indices[i] for i in dims_not_displayed)
        if any(type(id_) is int and id_ < 0 or id_ >= layer.data.shape[dim] for dim, id_ in zip(dims_not_displayed, idx)):
            return None, None
        # if not block and not self.slices_calculated[layer][dims_displayed][idx]:
        if not block and not self.feature_extractor.done_mask[idx]:
            raise ValueError("features not calculated for layer %s at %s" % (layer, idx))
        # while not self.slices_calculated[layer][dims_displayed][idx]:
        while not self.feature_extractor.done_mask[idx]:
            pass
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            idx = slice_indices
        return self.memmaps[0][idx], self.memmaps[1][idx]

    def clear_memmap(self):
        while len(self.memmaps) > 0:
            try:
                del self.memmaps[0]
            except:
                traceback.print_exc(file=sys.stdout)

    def remove_features(self, layer):
        if self.layer == layer:
            self.layer = None
            self.clear_memmap()
            self.feature_extractor.done_mask = None
        if layer not in self.prefix_map:
            return
        filename = self.prefix_map[layer]
        if layer in self.slices_calculated:
            del self.slices_calculated[layer]
        files = glob.glob(os.path.join(self.temp_folder, self.generate_filename(filename)))
        for file in files:
            os.remove(file)

    def init_file(self, layer, dims_displayed):
        self.layer = layer
        self.dims_displayed = dims_displayed
        self.clear_memmap()
        if layer in self.prefix_map:
            filename = self.prefix_map[layer]
        else:
            filename = self.random_prefix()
            self.prefix_map[layer] = filename
        if layer not in self.slices_calculated:
            self.slices_calculated[layer] = dict()
        if dims_displayed not in self.slices_calculated[layer]:
            self.slices_calculated[layer][dims_displayed] = np.zeros([layer.data.shape[i] for i in self.viewer.dims.not_displayed], bool)
        path_v = self.generate_filename(filename, dims_displayed, "_v")
        path_h = self.generate_filename(filename, dims_displayed, "_h")
        if layer.rgb:
            shape = layer.data.shape[:-1]
        else:
            shape = layer.data.shape
        if not os.path.exists(path_v):
            self.memmaps.append(np.memmap(path_v, shape=shape, dtype=float, mode="w+"))
            self.memmaps.append(np.memmap(path_h, shape=shape, dtype=float, mode="w+"))
            self.start_feature_calculation(layer)
        else:
            self.memmaps.append(np.memmap(path_v, shape=shape, dtype=float))
            self.memmaps.append(np.memmap(path_h, shape=shape, dtype=float))
            self.feature_extractor.done_mask = np.ones([shape[i] for i in layer_dims_not_displayed(layer)], bool)

    def generate_filename(self, prefix, dims_displayed=None, suffix=''):
        if dims_displayed is None:
            dims_displayed = "*"
        return os.path.join(self.temp_folder, "tmp_ftrs_%s_%s%s.dat" % (prefix, "_".join(str(d) for d in dims_displayed), suffix))

    def start_feature_calculation(self, layer):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            slice_indices = layer_slice_indices(layer)
            dims_not_displayed = layer_dims_not_displayed(layer)
        self.feature_extractor.start_jobs(layer.data, self.memmaps, slice_indices, dims_not_displayed, layer.rgb)

    @staticmethod
    def random_prefix():
        return ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(10))

    def clean(self):
        self.clear_memmap()
        if os.path.exists(self.temp_folder):
            shutil.rmtree(self.temp_folder)

    def clean_tmp(self):
        temp_dir = tempfile.gettempdir()
        temp_folders = glob.glob(os.path.join(temp_dir, "%s*%s" % (tempfile.gettempprefix(), TEMP_SUFFIX)))
        for fold in temp_folders:
            try:
                shutil.rmtree(fold)
            except PermissionError:
                print(f"Couldn't remove temp folder {fold}, skipping.")



from packaging import version
import napari


class StrComparableVersion(version.Version):
    def __init__(self, ver: version.Version):
        super().__init__(str(ver))

    def __eq__(self, other_ver):
        if type(other_ver) == str:
            other_ver = version.parse(other_ver)
        return super().__eq__(other_ver)

    def __gt__(self, other_ver):
        if type(other_ver) == str:
            other_ver = version.parse(other_ver)
        return super().__gt__(other_ver)

    def __ge__(self, other_ver):
        if type(other_ver) == str:
            other_ver = version.parse(other_ver)
        return super().__ge__(other_ver)

    def __lt__(self, other_ver):
        if type(other_ver) == str:
            other_ver = version.parse(other_ver)
        return super().__lt__(other_ver)

    def __le__(self, other_ver):
        if type(other_ver) == str:
            other_ver = version.parse(other_ver)
        return super().__le__(other_ver)


NAPARI_VERSION = StrComparableVersion(version.parse(napari.__version__))
__all__ = ["NAPARI_VERSION"]



from typing import List, Optional

from magicclass.widgets import CollapsibleContainer
from qtpy.QtWidgets import QWidget, QPushButton, QVBoxLayout, QSizePolicy
from qtpy.QtCore import Signal


class CollapsibleWidget(QWidget):
    expanded = Signal()
    collapsed = Signal()
    expansion_changed = Signal(bool)
    def __init__(self, text="", parent=None):
        super().__init__(parent=parent)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self._collapsed = True
        self._text = text
        self.collapse_button = QPushButton(self._prefix() + text, parent=self)
        self.collapse_button.setSizePolicy(QSizePolicy.Expanding, self.collapse_button.sizePolicy().verticalPolicy())
        self.collapse_button.clicked.connect(lambda: self.setCollapsed(not self._collapsed))
        self.collapse_button.setStyleSheet("Text-align:left")
        layout.addWidget(self.collapse_button)

        self.content_widget = QWidget(parent=self)
        self.content_widget.setVisible(not self._collapsed)
        layout.addWidget(self.content_widget)
        super().setLayout(layout)

    def setLayout(self, QLayout):
        self.content_widget.setLayout(QLayout)

    def layout(self):
        return self.content_widget.layout()

    def setCollapsed(self, is_collapsed):
        prev_val = self.isCollapsed()
        self._collapsed = bool(is_collapsed)
        self.collapse_button.setText(self._prefix() + self._text)
        self.content_widget.setVisible(not self._collapsed)
        if prev_val != self.isCollapsed():
            if self.isCollapsed():
                self.collapsed.emit()
            else:
                self.expanded.emit()
            self.expansion_changed.emit(not self.isCollapsed())

    def isCollapsed(self):
        return self._collapsed

    def collapse(self):
        self.setCollapsed(True)

    def expand(self):
        self.setCollapsed(False)

    def _prefix(self):
        return "\N{BLACK MEDIUM RIGHT-POINTING TRIANGLE} " if self._collapsed else "\N{BLACK MEDIUM DOWN-POINTING TRIANGLE} "


class CollapsibleWidgetGroup:
    def __init__(self, widget_list: Optional[List[CollapsibleWidget]] = None):
        self._widget_list = []
        self._handlers = dict()
        if widget_list is None:
            widget_list = []
        for widget in widget_list:
            if type(widget) != CollapsibleWidget:
                raise TypeError("%s is not CollapsibleWidget" % str(widget))
            self.addItem(widget)

    def addItem(self, widget: CollapsibleWidget):
        if widget in self._widget_list:
            raise ValueError("%s widget already in group" % str(widget))
        self._widget_list.append(widget)
        self._handlers[widget] = self._widget_expanded_handler(widget)
        widget.expanded.connect(self._handlers[widget])

    def removeItem(self, widget: CollapsibleWidget):
        if widget not in self._widget_list:
            return
        self._widget_list.remove(widget)
        widget.expanded.disconnect(self._handlers[widget])
        del self._handlers[widget]

    def _widget_expanded_handler(self, widget: CollapsibleWidget):
        def handler():
            for w2 in self._widget_list:
                if w2 != widget:
                    w2.collapse()
        return handler


class CollapsibleContainerGroup:
    def __init__(self, container_list: Optional[List[CollapsibleContainer]] = None):
        self._container_list = []
        self._handlers = dict()
        if container_list is None:
            container_list = []
        for container in container_list:
            if type(container) != CollapsibleContainer:
                raise TypeError("%s is not CollapsibleContainer" % str(container))
            self.addItem(container)

    def addItem(self, container: CollapsibleContainer):
        if container in self._container_list:
            return
        self._container_list.append(container)
        self._handlers[container] = self._container_expanded_handler(container)
        container._widget._expand_btn.clicked.connect(self._handlers[container])

    def removeItem(self, container: CollapsibleContainer):
        if container not in self._container_list:
            return
        self._container_list.remove(container)
        container._widget._expand_btn.clicked.disconnect(self._handlers[container])
        del self._handlers[container]

    def _container_expanded_handler(self, container: CollapsibleContainer):
        def handler(state: bool):
            if not state:
                return
            for c2 in self._container_list:
                if c2 != container:
                    c2._widget._collapse()
                    c2._widget._expand_btn.setChecked(False)
        return handler

    def __iter__(self):
        return iter(self._container_list)

def correct_container_size(container):
    if not container.collapsed:
        container.collapsed = True
        container.collapsed = False


import threading
import time
import traceback
from functools import wraps

import numpy as np
from magicgui._util import debounce
from napari.qt.threading import thread_worker
from napari._vispy.overlays.base import LayerOverlayMixin, VispySceneOverlay
from napari._vispy.utils import visual
from napari.layers import Labels
from vispy.scene import Polygon, Compound

from .interpolation_overlay import InterpolationOverlay, mutex as overlay_mutex


def execute_last(function=None, delay=0.2):

    def decorator(fn):
        from threading import Timer

        _store: dict = {"retry_worker": None, "mutex": threading.Lock()}

        @wraps(fn)
        def delayed(*args, **kwargs):
            mutex = _store["mutex"]
            def call_it(*args, **kwargs):
                _store["retry_worker"] = None
                with mutex:
                    print("calling function", flush=True)
                    fn(*args, **kwargs)
                    print("called function", flush=True)

            @thread_worker
            def retry(*args, **kwargs):
                while True:
                    if not mutex.locked():
                        call_it(*args, **kwargs)
                        return
                    time.sleep(delay)
                    yield
            if _store["retry_worker"] is not None:
                _store["retry_worker"].quit()
            _store["retry_worker"] = retry(*args, **kwargs)
            _store["retry_worker"].start()
            return None

        return delayed

    return decorator if function is None else decorator(function)


class VispyInterpolationOverlay(LayerOverlayMixin, VispySceneOverlay):
    layer: Labels

    def __init__(
        self, *, layer: Labels, overlay: InterpolationOverlay, parent=None
    ):
        points = np.asarray([(0, 0), (1, 1)])
        self._polygons = [Polygon(
            pos=points,
            border_method='gl',
            border_width=2
        ) for _ in range(1)]

        super().__init__(
            node=Compound(self._polygons),
            layer=layer,
            overlay=overlay,
            parent=parent,
        )
        # self.overlay.events.points_per_slice.connect(self._on_points_change)
        self.overlay.events.contour.connect(self._on_points_change)
        self.overlay.events.enabled.connect(self._on_enabled_change)

        layer.events.selected_label.connect(self._update_color)
        layer.events.colormap.connect(self._update_color)
        layer.events.opacity.connect(self._update_color)


        self.reset()
        self._update_color()
        # If there are no points, it won't be visible
        self.overlay.visible = True

    def _on_enabled_change(self):
        if self.overlay.enabled:
            self._on_points_change()

    # @execute_last
    def _on_points_change(self):
        print("on_points_change")
        # print("initial checks in points change:", time.time() - start, flush=True)
        contours = self.overlay.contour.copy()
        n_contours = len(contours)
        n_poly = len(self._polygons)
        if n_poly < n_contours:
            self._polygons.extend([Polygon(
                pos=np.asarray([(0, 0), (1, 1)]),
                border_method='agg',
                border_width=3
            ) for _ in range(n_contours - n_poly)])
        # print(f"extending polygons: {time.time() - start} (n_poly: {n_poly}, n_contours: {n_contours})", flush=True)
        for i, poly in enumerate(self._polygons):
            points = contours[i] if i < n_contours else None
            if points is not None and len(points) > 2:
                poly.visible = True
                poly.pos = points
            else:
                print("setting poly invisible", flush=True)
                poly.visible = False

    def _set_color(self, color):
        print("set_color")
        border_color = tuple(color[:3]) + (1,)  # always opaque
        polygon_color = color

        # Clean up polygon faces before making it transparent, otherwise
        # it keeps the previous visualization of the polygon without cleaning
        for poly in self._polygons:
            if polygon_color[-1] == 0:
                poly.mesh.set_data(faces=[])
            poly.color = polygon_color

            poly.border_color = border_color

    def _update_color(self):
        print("_update_color")
        layer = self.layer
        if layer._selected_label == layer.colormap.background_value:
            self._set_color((1, 0, 0, 0))
        else:
            self._set_color(
                layer._selected_color.tolist()[:3] + [layer.opacity]
            )

    @property
    def _dims_displayed(self):
        return self.layer._slice_input.displayed

    def reset(self):
        print("reset")
        super().reset()
        self._on_points_change()

visual.overlay_to_visual[InterpolationOverlay] = VispyInterpolationOverlay



from .interpolation_widget import InterpolationWidget
from .projections import SliceDisplayWidget
from .annotator_module import AnnotatorWidget
from .minimal_surface_widget import MinimalSurfaceWidget

import napari
from packaging import version

from .._napari_version import NAPARI_VERSION

if NAPARI_VERSION >= version.parse("0.4.15"):
    from .object_list import ListWidgetBB
    __all__ = ["InterpolationWidget", "SliceDisplayWidget", "AnnotatorWidget", "MinimalSurfaceWidget", "ListWidgetBB"]
else:
    if NAPARI_VERSION <= version.parse("0.4.12"):
        from napari.layers import Layer
        def data_to_world(self, position):
            """Convert from data coordinates to world coordinates.
            Parameters
            ----------
            position : tuple, list, 1D array
                Position in data coordinates. If longer then the
                number of dimensions of the layer, the later
                dimensions will be used.
            Returns
            -------
            tuple
                Position in world coordinates.
            """
            if len(position) >= self.ndim:
                coords = list(position[-self.ndim:])
            else:
                coords = [0] * (self.ndim - len(position)) + list(position)

            return tuple(self._transforms[1:].simplified(coords))
        Layer.data_to_world = data_to_world

    __all__ = ["InterpolationWidget", "SliceDisplayWidget", "AnnotatorWidget", "MinimalSurfaceWidget"]



from superqt.sliders import QDoubleRangeSlider


class QSymmetricDoubleRangeSlider(QDoubleRangeSlider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.valueChanged.connect(self.change_symmetrically)
        self._prev_value = self.value()

    def change_symmetrically(self, value):
        try:
            if value[0] != self._prev_value[0] and value[1] != self._prev_value[1]:
                return
            if value[0] != self._prev_value[0]:
                diff = value[0] - self._prev_value[0]
                value = (value[0], value[1]-diff)
            elif value[1] != self._prev_value[1]:
                diff = value[1] - self._prev_value[1]
                value = (value[0]-diff, value[1])
            self.valueChanged.disconnect(self.change_symmetrically)
            self.setValue(value)
            self.valueChanged.connect(self.change_symmetrically)
        finally:
            self._prev_value = value



import os.path
import warnings

import cv2
import napari
import numpy as np
from packaging import version

from qtpy.QtCore import QEvent, Qt, QObject
from qtpy.QtGui import QImage, QPixmap, QHideEvent, QShowEvent
from qtpy.QtWidgets import QWidget, QVBoxLayout, QLineEdit, QHBoxLayout, QLabel, QListWidgetItem, QListWidget, QMenu, \
    QComboBox, QPushButton, QStackedWidget, QDockWidget, QFileDialog, QSpinBox
from napari import Viewer
from napari.layers import Image, Labels
from napari.layers.labels._labels_constants import Mode
from scipy.ndimage import find_objects

from ._utils import WidgetWithLayerList
from .projections import SliceDisplayWidget
from .._napari_version import NAPARI_VERSION

if NAPARI_VERSION >= version.parse("0.4.15"):
    try:
        from napari_bbox import BoundingBoxLayer
    except ImportError:
        BoundingBoxLayer = None
else:
    BoundingBoxLayer = None

import itertools
import csv

import PIL.Image as PilImage

import time


def slice_len(slice_, count=None):
    if count is None:
        count = slice_.stop
    return len(range(*slice_.indices(count)))


if BoundingBoxLayer:
    class QObjectWidget(QWidget):
        def __init__(self, icon=None, name="Object", index=0, bounding_box=None, parent=None):
            super().__init__()
            self.textQVBoxLayout = QVBoxLayout()
            self.textUpQLabel = QLineEdit()
            self.textUpQLabel.setToolTip("Object name")
            self.id_label = QLabel()
            self.id_label.setText("Id: " + str(index))
            self.id_label.setToolTip("label index")
            self.textDownQLabel = QLabel()
            self.textQVBoxLayout.addWidget(self.textUpQLabel)
            self.textQVBoxLayout.addWidget(self.id_label)
            self.allQHBoxLayout  = QHBoxLayout()
            self.iconQLabel      = QLabel()
            self.allQHBoxLayout.addWidget(self.iconQLabel, 0)
            self.allQHBoxLayout.addLayout(self.textQVBoxLayout, 1)
            self.setLayout(self.allQHBoxLayout)
            if icon is not None:
                self.iconQLabel.setPixmap(icon)
            self.setTextUp(name)
            if bounding_box is not None:
                self.setTextDown(bounding_box)
            self.name = name
            self.parent = parent

        @property
        def name(self):
            return self.textUpQLabel.text()

        @name.setter
        def name(self, name):
            self.textUpQLabel.setText(name)

        def setTextUp(self, text):
            self.textUpQLabel.setText(text)

        def setTextDown(self, bb):
            self.textDownQLabel.setText(("corner 1: %s\ncorner 2: %s" % tuple(bb)))

        def setIcon(self, icon):
            self.iconQLabel.setPixmap(icon)
            self.iconQLabel.setFixedSize(icon.size())


    class QObjectListWidgetItem(QListWidgetItem):
        def __init__(self, name, bounding_box, index, parent, viewer, image_layer, mask_layer, channels_dim, *args):
            super().__init__(None, *args)
            if name is None:
                name = parent.name_template
            self.object_item = QObjectWidget(name=name, index=index, bounding_box=bounding_box, parent=self)
            self.setSizeHint(self.object_item.sizeHint())
            self.setFlags(self.flags() | Qt.ItemFlag.ItemIsEditable)
            self.parent = parent
            self.viewer = viewer
            self.image_layer = image_layer
            self.mask_layer = mask_layer
            self.bounding_box = bounding_box
            self.channels_dim = channels_dim

        def __hash__(self):
            return id(self)

        @property
        def idx(self):
            return int(float(self.object_item.id_label.text().split(":")[-1]))

        @idx.setter
        def idx(self, idx):
            self.object_item.id_label.setText("Id: " + str(idx))

        @property
        def icon(self):
            return self._icon

        @icon.setter
        def icon(self, icon):
            if len(icon) == 0:
                return
            self._icon = icon.reshape(icon.shape[0], icon.shape[1], -1)
            if self._icon.shape[-1] == 1:
                self._icon = np.tile(self._icon, (1, 1, 3))
            elif self._icon.shape[-1] == 4:
                self._icon = self._icon[..., :-1]
            scale = 64 / max(self._icon.shape[1], self._icon.shape[0])
            out_size = (int(self._icon.shape[1]*scale), int(self._icon.shape[0]*scale))
            self._resized_icon = cv2.resize(self._icon, out_size, interpolation=cv2.INTER_NEAREST_EXACT)
            img = QImage(self._resized_icon, self._resized_icon.shape[1], self._resized_icon.shape[0], self._resized_icon.shape[1] * 3, QImage.Format.Format_RGB888)
            self.pixmap = QPixmap()
            self.pixmap = self.pixmap.fromImage(img, Qt.ImageConversionFlag.ColorOnly)
            self.object_item.setIcon(self.pixmap)
            # self.update_icon()

        @property
        def bounding_box(self):
            return self._bounding_box

        @bounding_box.setter
        def bounding_box(self, bounding_box):
            self._bounding_box = bounding_box
            self.object_item.setTextDown(bounding_box)
            self.object_item.update()
            if self.parent.crop_image_layer is not None and \
                self.parent.selected_idx == self.parent.indexFromItem(self).row():
                self.parent.crop_image_layer.data = self.image_layer.data[self.bbox_idx]
                self.parent.crop_image_layer.translate = tuple(self.bounding_box[0, :])
                self.parent.crop_image_layer.refresh()
                self.parent.crop_mask_layer.data = self.mask_layer.data[self.bbox_idx]
                self.parent.crop_mask_layer.translate = tuple(self.bounding_box[0, :])
                self.parent.crop_mask_layer.refresh()
            self.update_icon()

        @property
        def bbox_idx(self):
            bbox_idx = tuple(
                slice(self.bounding_box[0, d], self.bounding_box[1, d]+1) for d in range(self.bounding_box.shape[1]))
            return bbox_idx

        @property
        def name(self):
            return self.object_item.name

        @name.setter
        def name(self, name):
            self.object_item.name = name

        def create_layers(self):
            if self.image_layer is not None:
                image_layer = Image(self.image_layer.data[self.bbox_idx],
                                    name="Cropped %s %d" % (self.image_layer.name, self.idx),
                                    colormap=self.image_layer.colormap, rgb=self.image_layer.rgb)
                image_layer.translate = self.bounding_box[0, :]+self.image_layer.translate
                self.viewer.add_layer(image_layer)
            else:
                image_layer = None
            if self.mask_layer is not None:
                mask_layer = Labels(self.mask_layer.data[self.bbox_idx],
                                    name="Cropped %s %d" % (self.mask_layer.name, self.idx))
                mask_layer.selected_label = self.idx
                mask_layer.translate = self.bounding_box[0, :]+self.mask_layer.translate
                mask_layer.brush_size = 1
                mask_layer.mode = Mode.PAINT
                self.mask_layer.events.data.connect(mask_layer.refresh)
                self.viewer.add_layer(mask_layer)
            else:
                mask_layer = None
            return image_layer, mask_layer

        def unselect(self):
            if self.parent.crop_image_layer is not None:
                if self.parent.crop_image_layer in self.viewer.layers:
                    self.viewer.layers.remove(self.parent.crop_image_layer)
                self.parent.crop_image_layer = None
            if self.parent.crop_mask_layer is not None:
                if self.parent.crop_mask_layer in self.viewer.layers:
                    self.viewer.layers.remove(self.parent.crop_mask_layer)
                if self.mask_layer is not None:
                    self.mask_layer.events.data.disconnect(self.parent.crop_mask_layer.refresh)
                self.parent.crop_mask_layer.mouse_drag_callbacks.remove(QObjectListWidgetItem.update_layer)
                self.parent.crop_mask_layer = None
            if self.parent.projections_widget is not None:
                self.viewer.window.remove_dock_widget(self.parent.projections_widget)
                self.parent.projections_widget = None
            if self.mask_layer is not None:
                self.mask_layer.visible = True
            if self.image_layer is not None:
                self.image_layer.visible = True

        def on_select(self):
            self.unselect()
            self.parent.crop_image_layer, self.parent.crop_mask_layer = self.create_layers()
            self.parent.crop_image_layer.name = self.parent.crop_image_layer.name
            self.parent.crop_mask_layer.name = self.parent.crop_mask_layer.name
            self.viewer.dims.set_point(
                np.arange(self.parent.crop_image_layer.ndim, dtype=int),
                np.mean(self.parent.crop_image_layer.extent.world, axis=0))
            self.parent.crop_mask_layer.events.set_data.connect(self.on_data_change)
            if self.image_layer.ndim > 2:
                self.parent.projections_widget = SliceDisplayWidget(self.viewer, self.parent.crop_image_layer, self.parent.crop_mask_layer, self.channels_dim)
                self.viewer.window.add_dock_widget(self.parent.projections_widget)
            self.parent.crop_mask_layer.mouse_drag_callbacks.append(self.update_layer)
            self.viewer.layers.selection.select_only(self.parent.crop_mask_layer)
            self.mask_layer.selected_label = self.idx
            self.mask_layer.visible = False
            self.image_layer.visible = False

        def on_data_change(self, event):
            if self.parent.crop_mask_layer is None:
                return
            self.mask_layer.data[self.bbox_idx] = self.parent.crop_mask_layer.data
            if self.parent.projections_widget is not None:
                self.parent.projections_widget.update_slider_ranges()
            self.parent.on_layer_event(event)

        @staticmethod
        def update_layer(layer, event):
            yield
            while event.type == "mouse_move":
                yield
            layer.refresh()

        def update_icon(self):
            visible_dims = list(self.viewer.dims.displayed)[-2:]
            bbox_idx = tuple(
                slice(max(self.bounding_box[0, d], 0), self.bounding_box[1, d]+1) if d in visible_dims
                else self.bounding_box[:, d].mean().astype(int)
                for d in range(self.bounding_box.shape[1])
            )
            if self.image_layer is None:
                icon = np.zeros([slice_len(bbox_idx[d]) for d in range(len(bbox_idx)) if type(bbox_idx[d]) is slice] + [3])
            else:
                icon = self.image_layer.data[bbox_idx]
            if self.image_layer is not None and not self.image_layer.rgb:
                max_ = self.image_layer.contrast_limits[1]
                min_ = self.image_layer.contrast_limits[0]
                icon = np.clip((icon.astype(float)-min_)/(max_ - min_), 0, 1)
                icon = (self.image_layer.colormap.map(icon.ravel()).reshape(icon.shape + (4,)) * 255).astype(np.uint8)[..., :3]
            else:
                icon = icon.astype(np.uint8)[..., :3]
            if self.mask_layer is not None:
                overlay = self.mask_layer.data[bbox_idx] == self.idx
                img = PilImage.fromarray(icon).convert(mode="RGBA")
                mask = PilImage.fromarray(overlay).convert(mode="RGBA")
                mask.putalpha(PilImage.fromarray((overlay*127).astype(np.uint8)))
                img.alpha_composite(mask)
                icon = np.asarray(img)
            self.icon = icon


    class ObjectListWidget(QListWidget):
        def __init__(self, viewer, name_template, bounding_box_layer, image_layer, mask_layer, channels_dim, indices=None):
            super().__init__()
            self._bounding_box_layer = None
            self._image_layer = None
            self._mask_layer = None
            self.crop_image_layer = None
            self.crop_mask_layer = None
            self.projections_widget = None
            self._mouse_down = False
            self.selected_idx = None
            self.previous_edge_color = None
            self._indices = None
            self.indices = indices
            self.viewer = viewer
            self.name_template = name_template
            self.channels_dim = channels_dim
            self.image_layer = image_layer
            self.mask_layer = mask_layer

            self.bounding_box_layer = bounding_box_layer
            self.setObjectName("Objects")
            self.update_items()
            self.itemClicked.connect(self.select_item)
            self.installEventFilter(self)
            self.viewer.bind_key('d', overwrite=True)(self.toggle_bb_visibility)
            # bounding_box_layer.events.set_data.connect(on_data_change)

        @property
        def bounding_box_layer(self):
            return self._bounding_box_layer

        @bounding_box_layer.setter
        def bounding_box_layer(self, new_layer):
            if self._bounding_box_layer == new_layer:
                self.update_items()
                return
            self._bounding_box_layer = new_layer
            if new_layer is not None:
                if self.bounding_box_change not in self._bounding_box_layer.mouse_drag_callbacks:
                    self._bounding_box_layer.mouse_drag_callbacks.append(self.bounding_box_change)
                if self._on_bb_double_click not in self._bounding_box_layer.mouse_double_click_callbacks:
                    self._bounding_box_layer.mouse_double_click_callbacks.append(self._on_bb_double_click)
                if "label" not in self._bounding_box_layer.features:
                    self._bounding_box_layer.features = {"label": np.arange(len(new_layer.data))}
                self._bounding_box_layer.current_properties |= {"label": 0}
                self._bounding_box_layer.refresh_text()
                if len(self._bounding_box_layer.data) > 0:
                    self._bounding_box_layer.text = {
                        "text": "{label:d}",
                        "size": 10,
                        "color": "green"
                    }
            self.update_items()

        @property
        def image_layer(self):
            return self._image_layer

        @image_layer.setter
        def image_layer(self, new_layer):
            self._image_layer = new_layer
            for i in range(self.count()):
                item = self.item(i)
                item.image_layer = new_layer
                item.update_icon()
            self.update_items()

        @property
        def mask_layer(self):
            return self._mask_layer

        @mask_layer.setter
        def mask_layer(self, new_layer):
            self._mask_layer = new_layer
            for i in range(self.count()):
                item = self.item(i)
                item.mask_layer = new_layer
                item.update_icon()
            self.update_items()

        @property
        def indices(self):
            return self._indices

        @indices.setter
        def indices(self, indices):
            if indices is None:
                indices = 1
            if type(indices) is int:
                starting_value = indices
                def index():
                    idx = starting_value
                    while True:
                        yield idx
                        idx += 1
                indices = index()
            self._indices = itertools.cycle(iter(indices))

        def on_layer_event(self, event):
            if event.type == "set_data" and not self._mouse_down:
                self.update_items(True)

        def select_item(self, item):
            selected_idx = self.indexFromItem(item).row()
            if self.mask_layer is not None and self.image_layer is not None:
                item.on_select()
            else:
                self.clearSelection()
                self.clearFocus()
                warnings.warn("Image and labels layer should be selected!")
                return
            if selected_idx == self.selected_idx or self.mask_layer is None or self.image_layer is None:
                self.selected_idx = None
                item.unselect()
                self.clearSelection()
                self.clearFocus()
                self.bounding_box_layer.edge_color[selected_idx] = self.previous_edge_color
                self.bounding_box_layer.data = self.bounding_box_layer.data
            else:
                if self.selected_idx is not None:
                    self.bounding_box_layer.edge_color[self.selected_idx] = self.previous_edge_color
                self.selected_idx = selected_idx
                self.previous_edge_color = self.bounding_box_layer.edge_color[selected_idx].copy()
                self.bounding_box_layer.edge_color[selected_idx] = (1., 0., 0., 1.)
                self.bounding_box_layer.data = self.bounding_box_layer.data

        def bounding_box_corners(self):
            if self._bounding_box_layer is None or len(self._bounding_box_layer.data) == 0:
                return []
            b_data = np.round(np.asarray(self._bounding_box_layer.data)).astype(int)
            if self.image_layer is not None:
                b_data -= self.image_layer.translate.astype(int)
            return np.concatenate([b_data.min(1, keepdims=True), b_data.max(1, keepdims=True)], axis=1)

        def eventFilter(self, source: QObject, event: QEvent) -> bool:
            if event.type() == QEvent.Close:
                if self.crop_image_layer is not None:
                    self.viewer.layers.remove(self.crop_image_layer)
                    self.viewer.layers.remove(self.crop_mask_layer)
                    self.viewer.window.remove_dock_widget(self.projections_widget)
            elif event.type() == QEvent.ContextMenu and source is self:
                menu = QMenu()
                menu.addAction("Create layers")
                if menu.exec_(event.globalPos()):
                    item = self.itemAt(event.pos())
                    item.create_layers()
                return True
            elif event.type() == QEvent.Enter and type(source) == QObjectWidget:
                item = source.parent
                if item is None:
                    return False
                idx = self.indexFromItem(item).row()
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    self.bounding_box_layer._value = (idx, None)
                    self.bounding_box_layer._set_highlight()
                return True
            elif event.type() == QEvent.Leave and type(source) == QObjectWidget:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    self.bounding_box_layer._value = (None, None)
                    self.bounding_box_layer._set_highlight()
                return True
            return super().eventFilter(source, event)

        def hideEvent(self, a0: QHideEvent) -> None:
            self.viewer.bind_key('d')(None)
            super().hideEvent(a0)

        def toggle_bb_visibility(self, _=None):
            if self.bounding_box_layer is not None and self.bounding_box_layer.visible:
                mode = self.bounding_box_layer.mode
                self.bounding_box_layer.visible = False
                yield
                self.bounding_box_layer.visible = True
                self.bounding_box_layer.mode = mode

        def bounding_box_change(self, layer, event):
            previous_data = np.asarray([bb.copy() for bb in layer.data])
            self._mouse_down = True
            yield
            if len(layer.data)>len(previous_data):
                self.bounding_box_layer.features["label"].iat[-1] = self.next_index()
                text = dict(layer.text)
                if NAPARI_VERSION == "0.4.15":
                    del text["values"]
                text["text"] = "{label:d}"
                layer.text = text
                layer.refresh_text()
            while event.type == "mouse_move":
                yield
            self._mouse_down = False
            new_data = np.asarray(layer.data)
            if self.image_layer:
                im_size = self.image_layer.data.shape
                if self.image_layer.rgb:
                    im_size = im_size[:-1]
            else:
                im_size = (np.inf,)*new_data.shape[-1]
            if len(new_data) > 0 and (np.any(np.all(new_data > im_size, 1))\
                    or np.any(np.all(new_data < 0, 1))):
                layer.data = previous_data
            if len(new_data) > len(previous_data):
                layer.data[-1][:] = np.clip(layer.data[-1], 0, np.asarray(self.image_layer.data.shape) - 1)
                layer.data = layer.data
            self.update_items()
            if self.projections_widget is not None:
                for p in self.projections_widget.projections:
                    p.update_overlay()
                    p.update()

        def update_items(self, update_all_icons=False):
            bounding_boxes = self.bounding_box_corners()
            if len(bounding_boxes) == 0:
                self.clear()
            elif len(bounding_boxes) < self.count():
                previous_data = np.asarray(list(np.unique(self.item(i).bounding_box, axis=0) for i in range(self.count())))
                removed_idx = np.argwhere(np.all(np.any(~np.equal(previous_data[np.newaxis], bounding_boxes[:, np.newaxis]), (-2, -1)), axis=0))
                removed_idx = np.squeeze(removed_idx)
                if removed_idx.ndim == 0:
                    self.takeItem(removed_idx)
                else:
                    self.clear()
            for i, bb in enumerate(bounding_boxes):
                curr_item = self.item(i)
                if curr_item is not None:
                    if not np.allclose(curr_item.bounding_box, bb):
                        curr_item.bounding_box = bb
                    if curr_item.image_layer is not self.image_layer:
                        curr_item.image_layer = self.image_layer
                    if curr_item.idx != self.bounding_box_layer.features["label"][i]:
                        curr_item.idx = self.bounding_box_layer.features["label"][i]
                    if update_all_icons:
                        curr_item.update_icon()

                else:
                    new_item = QObjectListWidgetItem(self.name_template, bb, self.bounding_box_layer.features["label"][i], self, self.viewer, self.image_layer, self.mask_layer, self.channels_dim)
                    self.insertItem(i, new_item)

        def insertItem(self, index, item: QObjectListWidgetItem):
            super().insertItem(index, item)
            self.setItemWidget(item, item.object_item)

        def addItem(self, item: QObjectListWidgetItem):
            self.insertItem(self.count(), item)

        def next_index(self):
            return next(self.indices)

        def _on_bb_double_click(self, layer=None, event=None):
            if len(self.bounding_box_layer.selected_data) == 0:
                return
            idx = next(iter(self.bounding_box_layer.selected_data))
            item = self.item(idx)
            self.setCurrentItem(item)
            self.select_item(item)


    class ListWidgetBB(WidgetWithLayerList):
        def __init__(self, viewer: Viewer):
            super().__init__(viewer, [("bounding_box", BoundingBoxLayer), ("image", Image), ("labels", Labels)], add_layers=False)
            layout = QVBoxLayout()
            self.viewer = viewer
            self.prev_n_layers = len(viewer.layers)
            self.channels_dim = None
            self.list_widget = None
            self.reset_index()
            self.bounding_box.combobox.currentIndexChanged.connect(self.bb_index_change)
            self.image.combobox.currentIndexChanged.connect(self.img_index_change)
            self.labels.combobox.currentIndexChanged.connect(self.mask_index_change)
            layout.addWidget(self.bounding_box.combobox)
            layout.addWidget(self.image.combobox)
            layout.addWidget(self.labels.combobox)

            self.list_widget_container = QStackedWidget()
            no_data_widget = QWidget()
            no_data_layout = QVBoxLayout()
            no_data_layout.addWidget(QLabel("No bounding box layer selected"))
            add_bb_button = QPushButton("Create Bounding Box layer")
            add_bb_button.clicked.connect(self.create_bb_layer)
            no_data_layout.addWidget(add_bb_button)
            no_data_widget.setLayout(no_data_layout)
            self.list_widget_container.addWidget(no_data_widget)
            self.list_widget_container.setCurrentIndex(0)
            layout.addWidget(self.list_widget_container)
            buttons_widget = QWidget()
            buttons_layout = QHBoxLayout()
            import_button = QPushButton("Import")
            import_button.clicked.connect(self.import_bounding_boxes)
            export_button = QPushButton("Export")
            export_button.clicked.connect(self.export_bounding_boxes)
            from_labels_button = QPushButton("From labels")
            from_labels_button.clicked.connect(self.bounding_boxes_from_labels)
            buttons_layout.addWidget(import_button)
            buttons_layout.addWidget(export_button)
            buttons_layout.addWidget(from_labels_button)
            buttons_widget.setLayout(buttons_layout)
            layout.addWidget(buttons_widget)
            self.setLayout(layout)
            self.name_template = "Object"
            if self.bounding_box.layer is not None:
                self.create_list_widget()
            self.installEventFilter(self)

        def bb_index_change(self, _=None):
            if self.bounding_box.layer is None:
                self.reset_index()
                self.remove_list_widget()
                return
            if self.list_widget is None:
                self.create_list_widget()
            self.list_widget.bounding_box_layer = self.bounding_box.layer
            if "label" in self.bounding_box.layer.features and len(self.bounding_box.layer.features):
                self.reset_index(max(self.bounding_box.layer.features["label"]) + 1)
            else:
                self.reset_index()

        def mask_index_change(self, _=None):
            if self.list_widget is not None and self.list_widget.mask_layer is not self.labels.layer:
                self.list_widget.mask_layer = self.labels.layer

        def img_index_change(self, _=None):
            self.update_channels_dim()

        @property
        def image_layer(self):
            return self.image.layer

        @property
        def mask_layer(self):
            return self.labels.layer

        def create_bb_layer(self, _):
            if self.image_layer is not None:
                self.viewer.add_layer(
                    BoundingBoxLayer(ndim=self.image_layer.ndim, edge_color="green", face_color="transparent"))

        def update_channels_dim(self):
            if self.image_layer is None:
                return
            smallest_dim = np.argsort(self.image_layer.data.shape)[0]
            if self.image_layer.data.shape[smallest_dim] <= 3:
                self.channels_dim = smallest_dim
            else:
                self.channels_dim = None
            if self.list_widget is not None:
                self.list_widget.image_layer = self.image_layer
                self.list_widget.channels_dim = self.channels_dim

        def create_list_widget(self):
            self.list_widget = ObjectListWidget(self.viewer, self.name_template, self.bounding_box.layer, self.image_layer, self.labels.layer, self.channels_dim, 1)
            self.list_widget_container.addWidget(self.list_widget)
            self.list_widget_container.setCurrentIndex(1)

        def remove_list_widget(self):
            while self.list_widget_container.count() > 1:
                self.list_widget_container.removeWidget(self.list_widget_container.widget(1))
            self.list_widget_container.setCurrentIndex(0)
            self.list_widget = None

        def export_bounding_boxes(self):
            if self.bounding_box.layer is None:
                return
            filename = QFileDialog.getSaveFileName(self, "Select...", None, "(*.txt)")[0]
            bboxes = np.asarray(list(map(lambda bb: np.concatenate([bb.min(axis=0), bb.max(axis=0)]), self.bounding_box.layer.data)))
            with open(filename, "w") as f:
                writer = csv.writer(f)
                for i, bb in enumerate(bboxes):
                    item = self.list_widget.item(i)
                    writer.writerow([item.name, item.idx] + list(bb))

        def import_bounding_boxes(self):
            filename = QFileDialog.getOpenFileName(self, "Select...", None, "(*.txt)")[0]
            if not filename:
                return
            names = []
            idxs = []
            data = []
            with open(filename, "r") as f:
                reader = csv.reader(f)
                for line in reader:
                    if len(line) == 0:
                        continue
                    names.append(line[0])
                    idxs.append(int(line[1]))
                    data.append(list(map(lambda x: float(x), line[2:])))
            data = np.asarray(data)
            bounding_box_corners = np.reshape(data, (len(data), 2, -1))
            mask = np.asarray(list(itertools.product((False, True), repeat=bounding_box_corners[0].shape[1])))
            bounding_boxes = np.asarray([np.where(mask, bbc[1], bbc[0]) for bbc in bounding_box_corners])
            bounding_box_layer = BoundingBoxLayer(bounding_boxes, name=os.path.basename(filename), edge_color="green", face_color="transparent", features={"label": np.asarray(idxs, dtype=int)})
            self.viewer.add_layer(bounding_box_layer)
            self.bounding_box.layer = bounding_box_layer

            for i, (name, idx) in enumerate(zip(names, idxs)):
                item = self.list_widget.item(i)
                item.name = name
                item.idx = idx
            self.reset_index(max(idxs) + 1)

        def bounding_boxes_from_labels(self):
            if self.labels.layer is None:
                return
            bb_layer = BoundingBoxLayer(ndim=self.image_layer.ndim, edge_color="green", face_color="transparent")
            bb_corners = find_objects(self.labels.layer.data)
            ids = []
            bbs = []
            for i, bb in enumerate(bb_corners):
                if bb is None:
                    continue
                min_ = [slice_.start for slice_ in bb]
                max_ = [slice_.stop - 1 for slice_ in bb]
                bb = np.asarray(np.where(list(itertools.product((False, True), repeat=self.image_layer.ndim)), max_, min_))
                bbs.append(bb)
                ids.append(i+1)
            bb_layer.data = bbs
            bb_layer.features["label"] = np.asarray(ids, dtype=int)
            self.viewer.add_layer(bb_layer)
            self.bounding_box.layer = bb_layer

        def reset_index(self, starting_index=1):
            if self.list_widget is not None:
                self.list_widget.indices = starting_index
else:
    class ListWidgetBB(QWidget):
        def __init__(self, *, parent=None):
            super().__init__(parent)
            layout = QVBoxLayout()
            layout.addWidget(QLabel("In order to use this widget, install the 'napari-bbox' plugin."))
            self.setLayout(layout)




import os

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
interpolate_style_path = os.path.join(__location__, "interpolate_button.qss")
from napari.resources._icons import write_colorized_svgs, _theme_path
from napari.settings import get_settings

interpolate_icon_path = os.path.join(__location__, "interpolate.svg").replace("\\", "/")
settings = get_settings()
theme_name = settings.appearance.theme
out = _theme_path(theme_name)
write_colorized_svgs(
    out,
    svg_paths=[interpolate_icon_path],
    colors=[(theme_name, 'icon')],
    opacities=(0.5, 1),
    theme_override={'warning': 'warning', 'logo_silhouette': 'background'},
)



import sys
from setuptools import setup, Extension
import os
import numpy as np

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None

# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions

extra_compile_args = ["/std:c++17", "/openmp"] if sys.platform == "win32"\
    else ["-std=c++17"] if sys.platform == "darwin"\
    else ["-std=c++17", "-fopenmp"]
extra_link_args = [] if sys.platform in ["win32", "darwin"] else ["-lgomp"]
extensions = [
    Extension(
        "napari_nd_annotator.minimal_contour._eikonal_wrapper",
        ["src/napari_nd_annotator/minimal_contour/_eikonal_wrapper.pyx"],
        extra_compile_args=extra_compile_args, extra_link_args=extra_link_args, language="c++", include_dirs=[np.get_include()]
    ),
    Extension(
        "napari_nd_annotator.mean_contour._essentials",
        ["src/napari_nd_annotator/mean_contour/cEssentialscy.pyx"],
        language="c++", include_dirs=[np.get_include()]
    ),
    Extension(
        "napari_nd_annotator.mean_contour._contour",
        ["src/napari_nd_annotator/mean_contour/contourcy.pyx"],
        language="c++", include_dirs=[np.get_include()]
    ),
    Extension(
        "napari_nd_annotator.mean_contour._reconstruction",
        ["src/napari_nd_annotator/mean_contour/reconstructioncy.pyx"],
        language="c++", include_dirs=[np.get_include()]
    )
]

CYTHONIZE = cythonize is not None

if CYTHONIZE:
    compiler_directives = {"language_level": 3, "embedsignature": True}
    extensions = cythonize(extensions, compiler_directives=compiler_directives)
else:
    extensions = no_cythonize(extensions)

setup(
    ext_modules=extensions,
    include_dirs=[np.get_include()]
)



import napari
import numpy as np
from napari import layers
from packaging import version
import warnings

from ._napari_version import NAPARI_VERSION

try:
    from napari.layers.labels.labels import _coerce_indices_for_vectorization
except ImportError:
    import numpy.typing as npt
    import inspect


    def _arraylike_short_names(obj):
        """Yield all the short names of an array-like or its class."""
        type_ = type(obj) if not inspect.isclass(obj) else obj
        for base in type_.mro():
            yield f'{base.__module__.split(".", maxsplit=1)[0]}.{base.__name__}'


    def _is_array_type(array: npt.ArrayLike, type_name: str) -> bool:
        return type_name in _arraylike_short_names(array)


    def _coerce_indices_for_vectorization(array, indices: list) -> tuple:
        """Coerces indices so that they can be used for vectorized indexing in the given data array."""
        if _is_array_type(array, 'xarray.DataArray'):
            # Fix indexing for xarray if necessary
            # See http://xarray.pydata.org/en/stable/indexing.html#vectorized-indexing
            # for difference from indexing numpy
            try:
                import xarray as xr
            except ModuleNotFoundError:
                pass
            else:
                return tuple(xr.DataArray(i) for i in indices)
        return tuple(indices)


if NAPARI_VERSION < "0.4.18":
    def layer_dims_displayed(layer: layers.Layer):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return layer._dims_displayed

    def layer_dims_not_displayed(layer: layers.Layer):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return layer._dims_not_displayed

    def layer_ndisplay(layer: layers.Layer):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return layer._ndisplay

    def layer_dims_order(layer: layers.Layer):
        return layer._dims_order

else:
    def layer_dims_displayed(layer: layers.Layer):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return layer._slice_input.displayed

    def layer_dims_not_displayed(layer: layers.Layer):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return layer._slice_input.not_displayed

    def layer_ndisplay(layer: layers.Layer):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return layer._slice_input.ndisplay

    def layer_dims_order(layer: layers.Layer):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            return layer._slice_input.order


if NAPARI_VERSION < "0.5.0":
    def layer_slice_indices(layer: layers.Layer):
        return layer._slice_indices

    def layer_get_order(layer: layers.Layer):
        return layer._get_order()
else:
    from napari.utils.misc import reorder_after_dim_reduction


    def layer_slice_indices(layer: layers.Layer):
        return tuple(slice(None) if np.isnan(p) else int(p) for p in layer._data_slice.point)


    def layer_get_order(layer: layers.Layer):
        order = reorder_after_dim_reduction(layer._slice_input.displayed)
        if len(layer.data.shape) != layer.ndim:
            # if rgb need to keep the final axis fixed during the
            # transpose. The index of the final axis depends on how many
            # axes are displayed.
            return (*order, max(order) + 1)

        return order



import asyncio.locks
import threading
import time
from copy import deepcopy
from typing import Optional

import napari
import numpy as np
import cv2
import scipy.ndimage
import skimage.draw
from PyQt5.QtCore import QEvent
from PyQt5.QtGui import QCursor, QPalette
from magicclass import magicclass, field, vfield, bind_key, MagicTemplate
from magicgui._util import debounce
from napari._qt.layer_controls.qt_labels_controls import QtLabelsControls
from napari._qt.qt_resources import get_current_stylesheet
from napari._qt.widgets.qt_mode_buttons import QtModePushButton
from napari.layers import Labels
from napari.utils.action_manager import action_manager
from napari.qt.threading import thread_worker

from scipy.interpolate import interp1d
from qtpy.QtCore import QThread, QObject, Signal, Qt
from qtpy.QtSvg import QSvgWidget
from scipy.ndimage import distance_transform_edt
from skimage.measure import regionprops
from skimage.morphology import binary_erosion
from skimage.transform import SimilarityTransform, warp
from ._utils import ProgressWidget
from .resources import interpolate_style_path
from .._helper_functions import layer_slice_indices, layer_dims_not_displayed, _coerce_indices_for_vectorization, \
    layer_get_order, layer_dims_displayed, layer_dims_order
from ..mean_contour import settings
from ..mean_contour.meanContour import MeanThread
from ..mean_contour._contour import calcRpsvInterpolation
from ..mean_contour._reconstruction import reconstruct
from ..mean_contour._essentials import magnitude
from .interpolation_overlay.interpolation_overlay import InterpolationOverlay, ContourList, mutex as overlay_mutex
import warnings

DISTANCE_BASED = "Distance-based"
CONTOUR_BASED = "Contour-based"
RPSV = "RPSV"

import faulthandler
faulthandler.enable()

def contour_cv2_mask_uniform(mask, contoursize_max, correct_orientation = True):
    mask = mask.astype(np.uint8)
    contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
    areas = [cv2.contourArea(cnt) for cnt in contours]
    max_ind = np.argmax(areas)
    contour = np.squeeze(contours[max_ind])
    contour = np.reshape(contour, (-1, 2))
    contour = np.append(contour, contour[0].reshape((-1, 2)), axis=0)
    contour = contour.astype('float32')

    rows, cols = mask.shape
    delta = np.diff(contour, axis=0)
    s = [0]
    for d in delta:
        dl = s[-1] + np.linalg.norm(d)
        s.append(dl)

    if s[-1] == 0:
        s[-1] = 1

    s = np.array(s) / s[-1]
    fx = interp1d(s, contour[:, 0] / rows, kind='linear')
    fy = interp1d(s, contour[:, 1] / cols, kind='linear')
    S = np.linspace(0, 1, contoursize_max, endpoint=False)
    X = rows * fx(S)
    Y = cols * fy(S)

    contour = np.transpose(np.stack([X, Y])).astype(np.float32)

    contour = np.stack((contour[:, 1], contour[:, 0]), axis=-1)
    if correct_orientation:
        centroid = contour.mean(0)
        start_index = np.argmin(np.abs(np.arctan2(*(contour - centroid).T)))
        contour = np.roll(contour, -start_index, 0)
    return contour


def average_mask(m1, m2, w1, w2):
    m1 = m1.astype(np.uint16)
    m2 = m2.astype(np.uint16)
    im_center = np.asarray(m1.shape)/2
    centroid1 = np.asarray(regionprops(m1)[0].centroid)
    centroid2 = np.asarray(regionprops(m2)[0].centroid)
    transl_1 = np.flip(centroid1 - im_center)
    transl_2 = np.flip(centroid2 - im_center)
    tform1 = SimilarityTransform(translation=transl_1)
    tform2 = SimilarityTransform(translation=transl_2)
    m1_translated = warp(m1, tform1, preserve_range=True).astype(bool)
    m2_translated = warp(m2, tform2, preserve_range=True).astype(bool)
    dt1_translated = distance_transform_edt(binary_erosion(m1_translated)) - distance_transform_edt(~m1_translated)
    dt2_translated = distance_transform_edt(binary_erosion(m2_translated)) - distance_transform_edt(~m2_translated)
    average_dist_translated = (w1 * dt1_translated + w2 * dt2_translated) / (w1+w2)
    average_mask_translated = average_dist_translated > 0
    transl_avg = np.flip(im_center - (centroid1*w1+centroid2*w2)/(w1+w2))
    tform_avg = SimilarityTransform(translation=transl_avg)

    return warp(average_mask_translated, tform_avg)


class MoveLoadingIconEventFilter(QObject):
    def __init__(self, icon: QSvgWidget):
        super().__init__()
        self.icon = icon
        self._is_loading = False
        self._is_above_viewer = False

    def update(self):
        if self._is_loading and self._is_above_viewer:
            self.icon.show()
        else:
            self.icon.hide()

    @property
    def is_loading(self):
        return self._is_loading

    @is_loading.setter
    def is_loading(self, value):
        self._is_loading = value
        self.update()

    def eventFilter(self, src: QObject, event: QEvent):
        if src == self.icon and event.type() == QEvent.Enter:
            pos = QCursor.pos()
        elif event.type() == QEvent.Type.HoverMove:
            pos = src.mapToGlobal(event.pos())
        elif event.type() == QEvent.Leave:
            self._is_above_viewer = False
            return False
        elif event.type() == QEvent.Enter:
            self._is_above_viewer = True
            return False
        else:
            return False
        pos.setX(pos.x() + 10)
        pos.setY(pos.y() + 10)
        self.icon.move(pos)
        self.update()
        return False


class InterpolationWorker(QObject):
    done = Signal("PyQt_PyObject")
    progress = Signal(int)
    dimension: int
    n_contour_points: int
    data: np.ndarray
    method: str
    max_iterations: int
    layer: Labels

    def run(self):
        try:
            dimension = self.dimension
            n_contour_points = self.n_contour_points
            selected_label = self.layer.selected_label
            data = (self.layer.data == selected_label).astype(np.uint8)
            method = self.method
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                layer_slice_template = list(layer_slice_indices(self.layer))
            prev_cnt = None
            prev_layer = None
            prev_mask = None
            start = time.time()
            for i in range(data.shape[dimension]):
                self.progress.emit(i)
                layer_slice = layer_slice_template.copy()
                layer_slice[dimension] = i
                cur_mask = data[tuple(layer_slice)]
                cur_mask = cur_mask.astype(np.uint8)
                if cur_mask.max() == 0:
                    continue
                next_layer_slice = layer_slice.copy()
                next_layer_slice[dimension] = i + 1
                prev_layer_slice = layer_slice.copy()
                prev_layer_slice[dimension] = i - 1
                if i + 1 < data.shape[dimension] \
                        and np.any(data[tuple(prev_layer_slice)])\
                        and np.any(data[tuple(next_layer_slice)]):
                    continue
                cnt = contour_cv2_mask_uniform(cur_mask, n_contour_points)
                if prev_cnt is None:
                    continue
                if method == RPSV:
                    stgs = settings.Settings(max_iterations=self.max_iterations,
                                             n_points=n_contour_points)
                    rpsv_thread = MeanThread([prev_cnt, cnt], stgs)
                    rpsv_thread.run()
                for j in range(prev_layer + 1, i):
                    inter_layer_slice = layer_slice_template.copy()
                    inter_layer_slice[dimension] = j
                    prev_w = i - j
                    cur_w = j - prev_layer
                    weights = [prev_w, cur_w]
                    if method == RPSV:
                        contours = rpsv_thread.contours
                        regularMean = np.zeros_like(contours[0].lookup[contours[0].parameterization, :])
                        for j in range(2):
                            regularMean += contours[j].lookup[contours[j].parameterization, :] * weights[j]
                        regularMean /= np.sum(weights)
                        q_mean = calcRpsvInterpolation(contours, weights)
                        guessRayLengths = np.zeros(contours[0].lookup[contours[0].parameterization].shape[0])
                        for i_contour in range(2):
                            contourtmp = contours[i_contour].lookup[contours[i_contour].parameterization]
                            contourlengths = magnitude(contourtmp)
                            guessRayLengths += contourlengths * weights[i_contour]
                        guessRayLengths /= np.sum(weights)
                        guessRayLengths = magnitude(regularMean)

                        qraylengths = magnitude(q_mean)
                        qraylengths[qraylengths < 1e-99] = 1e-99

                        dirs = q_mean / qraylengths.reshape(qraylengths.shape[0], 1)
                        r_mean_lengths, costs = reconstruct(q_mean, guessRayLengths.copy(), stgs, rpsv_thread.rpSignal)
                        mean_cnt = dirs * r_mean_lengths.reshape(r_mean_lengths.shape[0], 1)
                        mean_cnt = mean_cnt.astype(np.int32)
                        mask = np.zeros_like(data[tuple(inter_layer_slice)])
                        cv2.drawContours(mask, [np.flip(mean_cnt, -1)], 0, int(selected_label), -1)
                    elif method == CONTOUR_BASED:
                        mean_cnt = (prev_w * prev_cnt + cur_w * cnt)/(prev_w + cur_w)
                        mean_cnt = mean_cnt.astype(np.int32)
                        mask = np.zeros_like(data[tuple(inter_layer_slice)])
                        cv2.drawContours(mask, [np.flip(mean_cnt, -1)], 0, int(selected_label), -1)
                    elif method == DISTANCE_BASED:
                        mask = average_mask(prev_mask, cur_mask, prev_w, cur_w)
                        mask = mask.astype(np.uint8)
                        mask[mask > 0] = selected_label
                    else:
                        raise ValueError("method should be one of %s" % ((RPSV, CONTOUR_BASED, DISTANCE_BASED),))
                    data[tuple(inter_layer_slice)] = mask
                prev_cnt = cnt
                prev_layer = i
                prev_mask = cur_mask
            napari.notification_manager.receive_info("Done in %.4f s" % (time.time() - start))
            self.done.emit(data)
        except Exception as e:
            self.done.emit(None)
            raise e


@magicclass(name="Interpolation")
class InterpolationWidget(MagicTemplate):
    method = vfield(str,
                    widget_type="ComboBox",
                    options={
                        "choices": [CONTOUR_BASED, DISTANCE_BASED, RPSV],
                        "tooltip": "Interpolation method\n"
                                   f"{CONTOUR_BASED}: arithmetic mean of contour points\n"
                                   f"{DISTANCE_BASED}: interpolation between distance maps\n"
                                   f"{RPSV}: shape-aware mean of contours"
                    }
                    )
    n_contour_points = vfield(int,
                              widget_type="Slider",
                              name="Contour resolution",
                              options={
                                  "min": 10,
                                  "max": 1000,
                                  "value": 300,
                                  "tooltip": f"Number of contour points sampled. Used only for\"{RPSV}\" and \"{CONTOUR_BASED}\" methods"
                              })
    rpsv_max_iterations = field(int,
                                widget_type="SpinBox",
                                options={
                                    "min": 1,
                                    "max": 100,
                                    "value": 20,
                                    "tooltip": f"Maximum number of iterations for {RPSV}. Can be fewer if points converge."
                                })

    def __init__(self):
        self.progress_dialog = ProgressWidget(self.native, message="Interpolating slices...")
        # self.interpolation_worker.done.connect(lambda _: self.progress_dialog.setVisible(False))
        self._active_labels_layer = None
        self._is_painting = False
        self._labels_update_tmp = None
        self.workers = []
        self.workers_mutex = threading.Lock()
        self.overlay_worker = None
        self._stored_contours = []
        self.contour_mutex = overlay_mutex
        # layout.addStretch() -> how?

    def __post_init__(self):
        self._on_method_changed(self.method)
        self.native.layout().addStretch()

    def _initialize(self, viewer: napari.Viewer, annotator_widget):
        self._viewer = viewer
        self._on_active_layer_changed()
        self.viewer.dims.events.ndisplay.connect(self._ndisplay_changed)
        # self.viewer.dims.events.ndisplay.connect(lambda _: self._reset_contours() if self.viewer.dims.ndisplay == 3)
        self.viewer.layers.selection.events.active.connect(self._on_active_layer_changed)
        action_manager.register_action("napari-nD-annotator:interpolate", self.Interpolate,
                                       "Interpolate missing slices (I)", None)
        self._annotator_widget = annotator_widget

        self.viewer.dims.events.current_step.connect(self._current_step_changed)
        self.viewer.dims.events.order.connect(self._dims_order_changed)
        self.loading_icon = QSvgWidget("C:\\Users\\User\\Downloads\\loading.svg", self._viewer.window._qt_window)
        self.loading_icon.setStyleSheet("background: transparent;")
        self.loading_icon.setAutoFillBackground(False)
        palette = self.loading_icon.palette()
        palette.setBrush(QPalette.Base, Qt.transparent)
        self.loading_icon.setPalette(palette)
        self.loading_icon.setAttribute(Qt.WA_OpaquePaintEvent, False)
        self.loading_icon.setFixedSize(20, 20)
        self.loading_icon.setWindowFlags(Qt.Window | Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)
        self.loading_icon.setAttribute(Qt.WA_ShowWithoutActivating)
        self.loading_icon.setAttribute(Qt.WA_TranslucentBackground)

        self.loading_icon.show()
        self.event_filter = MoveLoadingIconEventFilter(self.loading_icon)
        self.loading_icon.installEventFilter(self.event_filter)
        self.viewer.window._qt_window.installEventFilter(self.event_filter)
        self.native.setMouseTracking(True)

    @debounce
    def _current_step_changed(self, _=None):
        if self.active_labels_layer is None or "interpolation" not in self.active_labels_layer._overlays:
            return
        if self.overlay_worker is not None:
            self.overlay_worker.quit()
        self._update_overlay()

    def _dims_order_changed(self):
        self._interpolate_all()

    def _ndisplay_changed(self):
        is_2d = self.viewer.dims.ndisplay == 2
        self.interpolate_button.options.update(enabled=is_2d)
        if is_2d:
            self._interpolate_all()
        else:
            self._reset_contours()

    # @thread_worker
    def _update_overlay(self):
        if self._stored_contours is None:
            contour = []
        else:
            contour = deepcopy(self._stored_contours[self.current_slice])
        with self.contour_mutex:
            self.active_labels_layer._overlays["interpolation"].contour = ContourList(contour)

    @property
    def viewer(self):
        return self._viewer

    @property
    def current_slice(self):
        if self.dimension is None:
            return None
        return layer_slice_indices(self.active_labels_layer)[self.dimension]

    def Interpolate(self, _=None):
        if self.active_labels_layer is None or self.dimension is None:
            return
        # self.progress_dialog.setMaximum(self.active_labels_layer.data.shape[self.dimension])
        # self.progress_dialog.setVisible(True)
        worker = self._store_interpolation()
        def on_worker_finished():
            self.interpolate_button.enabled = True
            self.event_filter.is_loading = False
        worker.finished.connect(on_worker_finished)
        self.interpolate_button.enabled = False
        worker.start()
        self.event_filter.is_loading = True

    @thread_worker
    def _store_interpolation(self):
        if self.active_labels_layer is None or "interpolation" not in self.active_labels_layer._overlays:
            return
        for i in range(len(self._stored_contours)):
            with self.contour_mutex:
                slice_contours = self._stored_contours[i]
            slice_template = list(layer_slice_indices(self.active_labels_layer))
            cur_slice = slice_template.copy()
            cur_slice[self.dimension] = i
            cur_slice = tuple(cur_slice)
            for contour in slice_contours:
                mask_slice = self.active_labels_layer.data[cur_slice]
                rr, cc = skimage.draw.polygon(contour[:, 1], contour[:, 0], mask_slice.shape)
                mask_slice[rr, cc] = self.active_labels_layer.selected_label
        self._reset_contours()
        self.active_labels_layer.refresh()

    def _interpolate_all(self):
        if self.active_labels_layer is None or "interpolation" not in self.active_labels_layer._overlays:
            return
        slices_with_data = np.argwhere(np.any(
            self.active_labels_layer.data == self.active_labels_layer.selected_label,
            axis=tuple(layer_dims_displayed(self.active_labels_layer))
        ))
        self._reset_contours()
        slices_with_data = np.atleast_1d(np.squeeze(slices_with_data))
        for idx in slices_with_data:
            self._start_interpolation(idx, 1)
        self._update_overlay()

    @thread_worker
    def _interpolate_from_slice(self, idx, direction):
        assert direction in [-1, 1], "direction should be either -1 or 1"
        if self.active_labels_layer is None:
            return

        dimension = self.dimension
        n_contour_points = self.n_contour_points
        selected_label = self.active_labels_layer.selected_label
        data = (self.active_labels_layer.data == selected_label).astype(np.uint8)
        method = self.method
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            layer_slice_template = list(layer_slice_indices(self.active_labels_layer))
        cur_slice = layer_slice_template.copy()
        cur_slice[dimension] = idx
        cur_mask = data[tuple(cur_slice)]
        cur_mask = cur_mask.astype(np.uint8)
        if cur_mask.max() == 0:
            return
        curr_cnt = contour_cv2_mask_uniform(cur_mask, n_contour_points)
        start_idx = idx + direction
        end_idx = -1 if direction == -1 else data.shape[dimension]
        for i in range(start_idx, end_idx, direction):
            layer_slice = layer_slice_template.copy()
            layer_slice[dimension] = i
            next_mask = data[tuple(layer_slice)]
            next_mask = next_mask.astype(np.uint8)
            if next_mask.max() == 0:
                continue
            if abs(i-idx) == 1:
                break
            next_cnt = contour_cv2_mask_uniform(next_mask, n_contour_points)
            if method == RPSV:
                stgs = settings.Settings(max_iterations=self.rpsv_max_iterations.value,
                                         n_points=n_contour_points)
                rpsv_thread = MeanThread([curr_cnt, next_cnt], stgs)
                rpsv_thread.run()
            for j in range(start_idx, i, direction):
                inter_layer_slice = layer_slice_template.copy()
                inter_layer_slice[dimension] = j
                prev_w = abs(i - j)
                cur_w = abs(j - idx)
                weights = [prev_w, cur_w]
                if method == RPSV:
                    contours = rpsv_thread.contours
                    regularMean = np.zeros_like(contours[0].lookup[contours[0].parameterization, :])
                    for k in range(2):
                        regularMean += contours[k].lookup[contours[k].parameterization, :] * weights[k]
                    regularMean /= np.sum(weights)
                    q_mean = calcRpsvInterpolation(contours, weights)
                    guessRayLengths = np.zeros(contours[0].lookup[contours[0].parameterization].shape[0])
                    for i_contour in range(2):
                        contourtmp = contours[i_contour].lookup[contours[i_contour].parameterization]
                        contourlengths = magnitude(contourtmp)
                        guessRayLengths += contourlengths * weights[i_contour]
                    guessRayLengths /= np.sum(weights)
                    guessRayLengths = magnitude(regularMean)

                    qraylengths = magnitude(q_mean)
                    qraylengths[qraylengths < 1e-99] = 1e-99

                    dirs = q_mean / qraylengths.reshape(qraylengths.shape[0], 1)
                    r_mean_lengths, costs = reconstruct(q_mean, guessRayLengths.copy(), stgs, rpsv_thread.rpSignal)
                    mean_cnt = dirs * r_mean_lengths.reshape(r_mean_lengths.shape[0], 1)
                    mean_cnt = mean_cnt.astype(np.int32)
                    mean_cnt = [mean_cnt]
                    # mask = np.zeros_like(data[tuple(inter_layer_slice)])
                    # cv2.drawContours(mask, [np.flip(mean_cnt, -1)], 0, int(selected_label), -1)
                elif method == CONTOUR_BASED:
                    mean_cnt = (prev_w * curr_cnt + cur_w * next_cnt) / (prev_w + cur_w)
                    mean_cnt = mean_cnt.astype(np.int32)
                    mean_cnt = [mean_cnt]
                    # mask = np.zeros_like(data[tuple(inter_layer_slice)])
                    # cv2.drawContours(mask, [np.flip(mean_cnt, -1)], 0, int(selected_label), -1)
                elif method == DISTANCE_BASED:
                    mask = average_mask(cur_mask, next_mask, prev_w, cur_w)
                    mask, n_labels = scipy.ndimage.label(mask)
                    mean_cnt = []
                    for label in range(1, n_labels+1):
                        mean_cnt.append(contour_cv2_mask_uniform(mask == label, n_contour_points))
                    # mask[mask > 0] = selected_label
                else:
                    raise ValueError("method should be one of %s" % ((RPSV, CONTOUR_BASED, DISTANCE_BASED),))
                yield j, mean_cnt
            break
        with self.contour_mutex:
            self._stored_contours[idx] = []

    def _execute_interpolation(self, idx):
        for direction in [-1, 1]:
            self._start_interpolation(idx, direction)

    def _start_interpolation(self, idx, direction):
        worker = self._interpolate_from_slice(idx, direction)
        with self.workers_mutex:
            self.workers.append(worker)
        worker.yielded.connect(self._store_contour)
        def remove_worker():
            with self.workers_mutex:
                self.workers[self.workers.index(worker)] = None
        worker.finished.connect(remove_worker)
        worker.start()

    @method.connect
    def _on_method_changed(self, new_method):
        self.rpsv_max_iterations.visible = new_method == RPSV
        self._interpolate_all()

    @property
    def active_labels_layer(self) -> Optional[Labels]:
        return self._active_labels_layer

    @active_labels_layer.setter
    def active_labels_layer(self, layer):
        if layer == self._active_labels_layer:
            return
        if self._active_labels_layer is not None and self._on_mouse_drag in self._active_labels_layer.mouse_drag_callbacks:
            self._active_labels_layer.mouse_drag_callbacks.remove(self._on_mouse_drag)
            self._active_labels_layer.events.labels_update.disconnect(self._on_labels_update)
            self._active_labels_layer.events.selected_label.disconnect(self._interpolate_all)
            if "interpolation" in self._active_labels_layer._overlays:
                self._active_labels_layer._overlays["interpolation"].contour = ContourList()
        self._active_labels_layer = layer if isinstance(layer, Labels) and layer.ndim == 3 else None
        if self._active_labels_layer is None or self._active_labels_layer.ndim < 3:
            self._reset_contours()
            return
        if self._on_mouse_drag not in self._active_labels_layer.mouse_drag_callbacks:
            self._active_labels_layer.mouse_drag_callbacks.append(self._on_mouse_drag)
            self._active_labels_layer.events.labels_update.connect(self._on_labels_update)
            self._active_labels_layer.events.selected_label.connect(self._interpolate_all)

        if "interpolation" not in self._active_labels_layer._overlays:
            interpolation_overlay = InterpolationOverlay()
            self._active_labels_layer._overlays.update({"interpolation": interpolation_overlay})
            labels_control: QtLabelsControls = self.viewer.window.qt_viewer.controls.widgets[self._active_labels_layer]
            if labels_control.button_grid.itemAtPosition(1, 1) is not None:
                return
            interpolate_button = QtModePushButton(
                self._active_labels_layer,
                'interpolate',
            )
            action_manager.bind_button(
                'napari-nD-annotator:interpolate', interpolate_button
            )
            interpolate_button.setStyleSheet(get_current_stylesheet([interpolate_style_path]))
            labels_control.button_group.addButton(interpolate_button)
            labels_control.interpolate_button = interpolate_button
            labels_control.button_grid.addWidget(labels_control.interpolate_button, 1, 1)
            self._active_labels_layer.bind_key("I", self.Interpolate)
        self._reset_contours()
        self._interpolate_all()

    @property
    def interpolate_button(self):
        for widget in self._list:
            if widget.name == "Interpolate":
                return widget

    def _enable_interpolation_button(self, *_):
        self.interpolate_button.enabled = True

    def _on_active_layer_changed(self, *_):
        self.active_labels_layer = self.viewer.layers.selection.active
        if self.active_labels_layer is None or self.active_labels_layer.ndim == 2:
            self.interpolate_button.enabled = False
            return
        self.interpolate_button.enabled = True

    @property
    def dimension(self):
        if self.active_labels_layer is None:
            return None
        dims_not_displayed = layer_dims_not_displayed(self.active_labels_layer)
        return None if len(dims_not_displayed) != 1 else dims_not_displayed[0]

    def _on_mouse_drag(self, layer: Labels, event):
        if (layer.mode not in ["erase", "paint"]
                and ("minimal_contour" not in layer._overlays or not layer._overlays["minimal_contour"].enabled)):
            return
        self._is_painting = True
        if layer.mode in ["erase", "paint"]:
            cur_slice = layer_slice_indices(layer)
            with self.contour_mutex:
                cnt_list = self._stored_contours[cur_slice[self.dimension]]
            if len(cnt_list) > 0:
                mask = np.zeros_like(layer.data[cur_slice])
                for i in range(len(cnt_list)):
                    cnt_list[i] = np.asarray(cnt_list[i], dtype=np.int32)

                cv2.drawContours(mask, cnt_list, -1, 1, -1)
                update_idx = np.nonzero(mask)
                order = layer_dims_order(layer)
                extended_idx = tuple(update_idx[i-1] if i in layer_dims_displayed(layer) else np.full_like(update_idx[0], cur_slice[order[i]]) for i in range(layer.ndim))
                extended_idx = _coerce_indices_for_vectorization(self.active_labels_layer.data, extended_idx)
                layer.data_setitem(extended_idx, layer.selected_label)
                with self.contour_mutex:
                    layer._overlays["interpolation"].contour = ContourList()


        yield
        while event.type == "mouse_move":
            yield
        self._is_painting = False
        if self._labels_update_tmp is not None:
            data, offset = self._labels_update_tmp
            self._labels_update_tmp = None
            if layer.mode == "erase" or not self._annotator_widget.autofill_objects:
                self._active_labels_layer.events.labels_update(data=data, offset=offset)
            cur_slice_data = layer.data[layer_slice_indices(layer)]
            if not np.any(cur_slice_data):
                self._interpolate_all()

    def _on_labels_update(self, event):
        if self._is_painting:
            self._labels_update_tmp = (event.data, event.offset)
            return
        self._execute_interpolation(layer_slice_indices(self._active_labels_layer)[self.dimension])

    def _on_set_data(self, event):
        ...

    def _store_contour(self, data):
        idx, contour_list = data
        uniques = np.empty(self.n_contour_points, dtype=bool)
        for i, contour in enumerate(contour_list):

            int_cnt = np.round(contour).astype(int)
            uniques[:-1] = np.any(int_cnt[:-1] != int_cnt[1:], axis=1)
            uniques[-1] = True
            contour = contour[uniques]
            contour = np.fliplr(contour)
            contour_list[i] = contour
        with self.contour_mutex:
            self._stored_contours[idx] = contour_list
        if idx == self.current_slice:
            with self.contour_mutex:
                self.active_labels_layer._overlays["interpolation"].contour = ContourList(contour_list)

    def _reset_contours(self):
        if self.dimension is None:
            with self.contour_mutex:
                self._stored_contours = None
            return
        with self.contour_mutex:
            self._stored_contours = [[] for _ in range(self.active_labels_layer.data.shape[self.dimension])]
            if self.active_labels_layer is not None:
                self.active_labels_layer._overlays["interpolation"].contour = ContourList()

    def _reset_interpolation_worker(self):
        self.interpolation_worker = None

    def _reset_overlay_worker(self):
        self.interpolation_worker = None


import numpy as np
from napari._pydantic_compat import Field
from napari.components.overlays import SceneOverlay
from napari.layers import Labels
import scipy

class MinimalContourOverlay(SceneOverlay):
    """Overlay that displays a polygon on a scene.

    This overlay was created for drawing polygons on Labels layers. It handles
    the following mouse events to update the overlay:
    - Mouse move: Continuously redraw the latest polygon point with the current
    mouse position.
    - Mouse press (left button): Adds the current mouse position as a new
    polygon point.
    - Mouse double click (left button): If there are at least three points in
    the polygon and the double-click position is within completion_radius
    from the first vertex, the polygon will be painted in the image using the
    current label.
    - Mouse press (right button): Removes the most recent polygon point from
    the list.

    Attributes
    ----------
    enabled : bool
        Controls whether the overlay is activated.
    points : list
        A list of (x, y) coordinates of the vertices of the polygon.
    use_double_click_completion_radius : bool
        Whether double-click to complete drawing the polygon requires being within
        completion_radius of the first point.
    completion_radius : int | float
        Defines the radius from the first polygon vertex within which
        the drawing process can be completed by a left double-click.
    """

    enabled: bool = False
    anchor_points: list = Field(default_factory=list)
    last_anchor_to_current_pos_contour: list = Field(default_factory=list)
    current_pos_to_first_anchor_contour: list = Field(default_factory=list)
    stored_contour: list = Field(default_factory=list)
    use_straight_lines: bool = False
    contour_smoothness: float = 1.
    contour_width: int = 3

    def add_polygon_to_labels(self, layer: Labels) -> None:
        if len(self.stored_contour) > 2:
            contour = np.asarray(self.stored_contour)
            if self.contour_smoothness < 1.:
                contour = self.smooth_contour(contour)
            layer.paint_polygon(np.round(contour), layer.selected_label)
        self.stored_contour = []
        self.last_anchor_to_current_pos_contour = []
        self.current_pos_to_first_anchor_contour = []
        self.anchor_points = []

    def smooth_contour(self, contour: np.ndarray):
        coefficients = max(3, round(self.contour_smoothness * len(contour)))
        mask_2d = ~np.all(contour == contour.min(), axis=0)
        points_2d = contour[:, mask_2d]
        center = points_2d.mean(0)
        points_2d = points_2d - center
        tformed = scipy.fft.rfft(points_2d, axis=0)
        tformed[0] = 0
        inv_tformed = scipy.fft.irfft(tformed[:coefficients], len(points_2d), axis=0) + center
        contour[:, mask_2d] = inv_tformed
        return contour


import numpy as np
from ._essentials import ReconstructionMethods, Contour
import matplotlib.pyplot as plt

# loads a contour from file contName with nPoi points and a resolution of nPoi*resMultiplier
def loadContour(contName, nPoi, resMultiplier):
    pts = np.genfromtxt(contName, delimiter=",")
    res = Contour(pts.copy(), nPoi, resMultiplier)
    return res

# plots a 1-variate function
def plotFunction(x, fx, title=""):
    plotfig = plt.figure()
    plt.title(title)
    plt.plot(x,fx)
    return plotfig


def plotContours(contours,labels=None,colors=None,title=""):
    plotfig = plt.figure()
    plt.title(title)
    for i in range(len(contours)):
        curr = contours[i]
        clabel = ""
        ccolor = "black"
        if labels:
            clabel = labels[i]
        if colors:
            ccolor = colors[i]
        plt.plot(curr[:,1],-1*curr[:,0], label=clabel, color=ccolor)
    plt.legend()
    return plotfig

def getReconMethod(reconText):
    switcher = {
        'Newton': ReconstructionMethods.NEWTON,
        'Gradient descent': ReconstructionMethods.GRADIENT_DESCENT,
        'Conjugate gradient': ReconstructionMethods.CG,
        'Skip reconstruction': ReconstructionMethods.SKIP,
        'Jozsi gradient': ReconstructionMethods.JOZSI_GRADIENT
    }
    return switcher.get(reconText)


from pathlib import Path
from typing import Optional, Union
import napari
from magicgui.widgets import Container, FunctionGui, create_widget
from qtpy.QtWidgets import QWidget, QComboBox, QVBoxLayout, QSizePolicy, QScrollArea
from qtpy.QtCore import Qt
from keyword import iskeyword
from abc import ABCMeta
from .persistence import PersistentWidget


class PostInitCaller(ABCMeta):
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        try:
            obj.__post_init__(*args, **kwargs)
        except AttributeError:
            raise AttributeError("Class should define function '__post_init__()")
        return obj


class WidgetWithLayerList2(Container, metaclass=PostInitCaller):
    def __init__(self, viewer: napari.Viewer, layers, add_layers=True, scrollable=False, persist=True, **kwargs):
        super().__init__(scrollable=scrollable,)
        self.viewer = viewer
        self.layers = dict()  # str -> QComboBox
        self.persist = persist
        for layer in layers:  # every layer should be a tuple: (layer_name, layer_type) or (layer_name, layer_type, layer_display_name)
            if len(layer) == 2:
                layer_name, layer_type = layer
                layer_displayed_name = layer_name.replace("_", " ")
            elif len(layer) == 3:
                layer_name, layer_type, layer_displayed_name = layer
            else:
                raise IndexError("every layer should be a tuple: (layer_name, layer_type)"
                                 " or (layer_name, layer_type, layer_display_name)")
            if not layer_name.isidentifier() or iskeyword(layer_name):
                raise ValueError("layer name '%s' is not a valid attribute name (cannot be accessed as 'obj.%s')" % (layer_name, layer_name))
            new_widget = create_widget(name=layer_name, label=layer_displayed_name, annotation=layer_type)
            self.__setattr__(layer_name, new_widget)
            self.layers[layer_name] = new_widget
        if add_layers:
            self.extend(self.layers)
        self.changed.connect(self._on_change),

    def __post_init__(self, *_, **__):
        if self.persist:
            self._load(quiet=True)

    def _on_change(self) -> None:
        if self.persist:
            self._dump()

    @property
    def _dump_path(self) -> Path:
        from magicgui._util import user_cache_dir

        name = getattr(self.__class__, "__qualname__", str(self.__class__))
        name = name.replace("<", "-").replace(">", "-")  # e.g. <locals>
        return user_cache_dir() / f"{self.__class__.__module__}.{name}"

    def _dump(self, path: Optional[Union[str, Path]] = None) -> None:
        super()._dump(path or self._dump_path)

    def _load(self, path: Optional[Union[str, Path]] = None, quiet: bool = False) -> None:
        super()._load(path or self._dump_path, quiet=quiet)


class WidgetWithLayerList(PersistentWidget):
    def __init__(self, viewer: napari.Viewer, layers, persistence_id=None, add_layers=True, scrollable=True, **kwargs):
        super().__init__(persistence_id, **kwargs)
        layout = QVBoxLayout(self)
        layers_layout = QVBoxLayout()
        if scrollable:
            self.scroll_area = QScrollArea(self)
            self.scroll_area.setWidgetResizable(True)
            self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
            self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.central_widget = QWidget(self)
        self.viewer = viewer
        self.layers = dict()  # str -> QComboBox
        for layer in layers:  # every layer should be a tuple: (layer_name, layer_type) or (layer_name, layer_type, layer_display_name)
            if len(layer) == 2:
                layer_name, layer_type = layer
                layer_displayed_name = layer_name.replace("_", " ")
            elif len(layer) == 3:
                layer_name, layer_type, layer_displayed_name = layer
            else:
                raise IndexError("every layer should be a tuple: (layer_name, layer_type)"
                                 " or (layer_name, layer_type, layer_display_name)")
            if not layer_name.isidentifier() or iskeyword(layer_name):
                raise ValueError("layer name '%s' is not a valid attribute name (cannot be accessed as 'obj.%s')" % (layer_name, layer_name))
            self.layers[layer_name] = self.LayerRecord(self, layer_type, layer_displayed_name)
            if add_layers:
                layers_layout.addWidget(self.layers[layer_name].combobox)
        layout.addLayout(layers_layout)
        if scrollable:
            self.scroll_area.setWidget(self.central_widget)
            layout.addWidget(self.scroll_area)
        else:
            layout.addWidget(self.central_widget)

    def setLayout(self, a0: 'QLayout') -> None:
        self.central_widget.setLayout(a0)

    def __getattr__(self, item):
        if item in self.layers:
            return self.layers[item]
        raise AttributeError("No attribute named %s" % item)

    class LayerRecord:
        def __init__(self, parent, layer_type, display_name="Select a Layer"):
            self.parent = parent
            self.layer_type = layer_type
            self.combobox = QComboBox()
            self.display_name = display_name
            self.combobox.addItem("[%s]" % display_name)
            self.combobox.currentIndexChanged.connect(self.on_layer_index_change)
            self.combobox.setToolTip(display_name)
            self.combobox.setSizePolicy(QSizePolicy.Ignored, self.combobox.sizePolicy().verticalPolicy())
            self.viewer.layers.events.connect(self.on_layer_list_change)
            self._moved_layer = None
            self._layer_name = None
            self.on_layer_list_change()

        @property
        def viewer(self) -> napari.Viewer:
            return self.parent.viewer

        @property
        def layer(self):
            return self.viewer.layers[self._layer_name] if self._layer_name and self._layer_name in self.viewer.layers\
                else None

        @layer.setter
        def layer(self, layer):
            if layer is None:
                self._layer_name = None
                self.combobox.setCurrentIndex(0)
                return
            if isinstance(layer, self.layer_type):
                new_layer_name = layer.name
            elif type(layer) is str:
                new_layer_name = layer
            else:
                raise TypeError("layer should be %s or str" % self.layer_type)
            if new_layer_name == self._layer_name:
                return
            if new_layer_name in self.viewer.layers:
                self.combobox.setCurrentText(new_layer_name)
                self._layer_name = new_layer_name

        def on_layer_index_change(self, index):
            if index == 0:
                if self.combobox.count() > 1:
                    self.combobox.setCurrentText(self._layer_name)
                    return
            elif index > 0:
                self._layer_name = self.combobox.itemText(index)

        def on_layer_list_change(self, event=None):
            type_ = event.type if event else None
            if type_ not in ["moved", "inserted", "removed", "name", None]:
                return
            filtered = list(filter(lambda layer: isinstance(layer, self.layer_type), self.viewer.layers))
            if len(filtered) == 0 and type_ != "removed":
                return
            if type_ in ["moved", "inserted", "removed", None]:
                if self.combobox.count() == len(filtered) and \
                        all((layer.name == self.combobox.itemText(i+1) for i, layer in enumerate(filtered))):
                    return
                self.combobox.blockSignals(True)
                self.combobox.clear()
                self.combobox.addItem("[%s]" % self.display_name)
                for layer in filtered:
                    self.combobox.addItem(layer.name)
                    if layer.name == self._layer_name:
                        self.combobox.setCurrentText(layer.name)
                self.combobox.blockSignals(False)
                if self._layer_name != self.combobox.currentText():
                    self.combobox.currentTextChanged.emit(self.combobox.currentText())
                    self.combobox.currentIndexChanged.emit(self.combobox.currentIndex())
                if self.combobox.count() > 1 and (self.combobox.currentIndex() == 0 or self.layer not in filtered):
                    self.layer = filtered[0]
                elif self.combobox.count() == 1:
                    self.layer = None
            elif type_ == "name":
                self.combobox.blockSignals(True)
                for i in range(len(filtered)):
                    if self.combobox.itemText(i+1) != filtered[i].name:
                        if self.combobox.itemText(i+1) == self._layer_name:
                            self._layer_name = filtered[i].name
                        self.combobox.setItemText(i+1, filtered[i].name)
                        break
                self.combobox.blockSignals(False)



import itertools

import numpy as np
from PIL import Image as PILImage
from qtpy.QtCore import QSettings, Qt
from qtpy.QtGui import QImage, QPixmap, QResizeEvent, QColor
from qtpy.QtWidgets import QLabel, QSizePolicy, QWidget, QVBoxLayout, QSlider, QHBoxLayout, QGridLayout, QCheckBox, \
    QColorDialog, QPushButton, QWIDGETSIZE_MAX, QDockWidget
from napari.layers.labels._labels_constants import Mode
from qtpy import QtCore

from .._helper_functions import layer_dims_not_displayed, layer_slice_indices
from matplotlib import colors

class DataProjectionWidget(QLabel):
    def __init__(self, viewer, image_layer, mask_layer, displayed_axes, slices=None,
                 flip_image=False, crosshair_color=None):
        super().__init__()
        self._crosshair_color = None
        self.viewer = viewer
        self.image_layer = image_layer
        self.mask_layer = mask_layer
        self.flip_image = flip_image
        self.max_width, self.max_height = QWIDGETSIZE_MAX, QWIDGETSIZE_MAX
        self.slices = list(slices) or [s//2 for s in self.image_data.shape]
        self.displayed_axes = displayed_axes
        overlay_shape = [self.image_data.shape[i] for i in range(self.image_layer.ndim) if self.im_idx[i] == slice(None)][:2]
        self._overlay = np.zeros(overlay_shape, np.uint8)
        if crosshair_color is None:
            self.crosshair_color = (255, 0, 0, 160)
        else:
            self.crosshair_color = crosshair_color
        self.update()
        self.pixmap = None
        self.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.mousePressEvent = self.on_click
        self.setFrameStyle(QLabel.Raised | QLabel.Box)

    def setSlices(self, *args):
        assert len(args) % 2 == 0
        for i in range(0, len(args), 2):
            self.slices[args[i]] = max(min(args[i+1], self.image_data.shape[args[i]]-1), 0)
        self.update()

    @property
    def crosshair_color(self):
        return self._crosshair_color

    @crosshair_color.setter
    def crosshair_color(self, new_color):
        if all(map(lambda v: 0.<=v<=1., new_color)):
            new_color = map(lambda v: int(v*255), new_color)
        self._crosshair_color = tuple(new_color)
        self.update()

    @property
    def overlay(self):
        overlay = (np.ones(self._overlay.shape + (4,)) * self.crosshair_color).astype(np.uint8)
        overlay[..., -1] *= self._overlay
        if self.flip_image:
            overlay = np.transpose(overlay, (1, 0, 2))
        return PILImage.fromarray(overlay)

    @overlay.setter
    def overlay(self, new_overlay):
        self._overlay = new_overlay > 0

    def update_overlay(self, layer=None, coordinates=None):
        overlay_shape = [self.image_data.shape[i] for i in range(self.image_layer.ndim) if
                         self.im_idx[i] == slice(None)][:2]
        overlay = np.zeros(overlay_shape, bool)
        if layer is not None and coordinates is not None:
            if layer._mode in [Mode.ERASE, Mode.PAINT]:
                if all(layer_dim == proj_dim for layer_dim, proj_dim in
                       zip(sorted(self.viewer.dims.displayed), sorted(self.displayed_axes))):
                    if 0 <= coordinates[self.displayed_axes[0]] < overlay.shape[1 if self.flip_image else 0]:
                        if self.flip_image:
                            overlay[:, coordinates[self.displayed_axes[0]]] = True
                        else:
                            overlay[coordinates[self.displayed_axes[0]]] = True
                    if 0 <= coordinates[self.displayed_axes[1]] < overlay.shape[0 if self.flip_image else 1]:
                        if self.flip_image:
                            overlay[coordinates[self.displayed_axes[1]]] = True
                        else:
                            overlay[:, coordinates[self.displayed_axes[1]]] = True
                else:
                    for dim in self.viewer.dims.displayed:
                        if dim == self.displayed_axes[0]:
                            start = max(coordinates[dim] - layer.brush_size // 2, 0)
                            end = min(max(coordinates[dim] + layer.brush_size // 2 + 1, 0), overlay.shape[1 if self.flip_image else 0])
                            if self.flip_image and coordinates[self.displayed_axes[1]] < overlay.shape[0]:
                                overlay[coordinates[self.displayed_axes[1]], start:end] = True
                            elif coordinates[self.displayed_axes[1]] < overlay.shape[1]:
                                overlay[start:end, coordinates[self.displayed_axes[1]]] = True
                        elif dim == self.displayed_axes[1]:
                            start = max(coordinates[dim] - layer.brush_size // 2, 0)
                            end = min(max(coordinates[dim] + layer.brush_size // 2 + 1, 0), overlay.shape[0 if self.flip_image else 1])
                            if self.flip_image and coordinates[self.displayed_axes[0]] < overlay.shape[1]:
                                overlay[start:end, coordinates[self.displayed_axes[0]]] = True
                            elif coordinates[self.displayed_axes[0]] < overlay.shape[0]:
                                overlay[coordinates[self.displayed_axes[0]], start:end] = True
        elif bool(layer) ^ bool(coordinates): # XOR
            raise ValueError("layer and coordinates should be both provided or left None")
        self.overlay = overlay

    @property
    def image_data(self):
        return self.image_layer.data

    @property
    def mask_data(self):
        return self.mask_layer.data

    @property
    def im_idx(self):
        return tuple(
                self.slices[i] if i not in self.displayed_axes
                    else slice(None) for i in range(self.image_layer.ndim))

    def update(self, update_icon=True, new_size=None):
        if update_icon:
            if any(self.im_idx[i] != slice(None) and self.im_idx[i] >= self.image_data.shape[i]
                    for i in range(self.image_layer.ndim)):
                im_shape = tuple(self.image_data.shape[i] for i in range(self.image_layer.ndim)
                                 if self.im_idx[i] == slice(None))[:2]
                im = np.zeros(im_shape + (4,), np.uint8)
                mask = np.zeros_like(im)
            else:
                if self.image_layer.rgb:
                    im = self.image_data[self.im_idx]
                else:
                    im = self.image_data[self.im_idx]
                    max_ = self.image_layer.contrast_limits[1]
                    min_ = self.image_layer.contrast_limits[0]
                    im = np.clip((im-min_)/(max_-min_), 0, 1)
                    im = (self.image_layer.colormap.map(im.ravel()).reshape(im.shape + (4,))*255).astype(np.uint8)[..., :3]

                mask = (self.mask_data[self.im_idx] > 0).astype(np.uint8)
                alpha = mask * 180
                mask = np.tile(mask[..., np.newaxis], (1, 1, 4))*255

                mask[..., -1] = alpha

                if self.flip_image:
                    im = np.transpose(im, (1, 0, 2))
                    mask = np.transpose(mask, (1, 0, 2))
            im = PILImage.fromarray(im.astype(np.uint8)).convert(mode="RGBA")
            mask = PILImage.fromarray(mask).convert(mode="RGBA")
            icon = PILImage.alpha_composite(im, mask)
            icon = PILImage.alpha_composite(icon, self.overlay).convert("RGB")
            icon = np.asarray(icon)
            _icon = icon.reshape(icon.shape[0], icon.shape[1], -1)
            if _icon.shape[-1] == 1:
                _icon = np.tile(_icon, (1, 1, 3))
            elif _icon.shape[-1] == 4:
                _icon = _icon[..., :-1]
            self.img = QImage(_icon, _icon.shape[1], _icon.shape[0], _icon.shape[1]*3, QImage.Format.Format_RGB888)
        pixmap = QPixmap()
        pixmap = pixmap.fromImage(self.img, QtCore.Qt.ImageConversionFlag.ColorOnly)
        self.setMinimumSize(64, 64)
        if new_size is not None:
            w = new_size.width()
            h = new_size.height()
        else:
            w = self.size().width()
            h = self.size().height()
        pixmap = pixmap.scaled(w, h, QtCore.Qt.KeepAspectRatio)
        self.setPixmap(pixmap)

    def on_resize(self, new_size):
        self.update(update_icon=False, new_size=new_size)

    def resizeEvent(self, a0: QResizeEvent) -> None:
        self.on_resize(a0.size())

    def on_click(self, _):
        order = tuple(filter(lambda x: x not in self.displayed_axes, self.viewer.dims.order))
        order = order + tuple(self.displayed_axes)
        self.viewer.dims.order = order
        self.viewer.dims.current_step = tuple(self.viewer.dims.current_step[dim] if dim in self.displayed_axes
                                         else self.slices[dim]+self.image_layer.translate[dim]
                                         for dim in range(len(self.viewer.dims.current_step)))


class SliceDisplayWidget(QWidget):
    def __init__(self, viewer, image_layer, mask_layer, channels_dim=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setObjectName("Projections")
        main_layout = QVBoxLayout()
        self.settings = QSettings()
        self.overlay_color = self.settings.value("projection_overlay_color", (255, 0, 0, 160))
        self.slider_labels = []
        self.sliders = []
        self.slider_widgets = []
        self.offset = image_layer.translate
        self.viewer = viewer
        self.image_layer = image_layer
        self.shown = False
        for dim in range(viewer.dims.ndim):
            label = QLabel("%d" % dim)
            slider = QSlider(QtCore.Qt.Horizontal)
            slider.setMinimum(0)
            slider.setMaximum(image_layer.data.shape[dim])
            slider.setSingleStep(1)
            slider.setFixedWidth(100)
            slider.valueChanged.connect(self.slider_callback(dim))
            slider_layout = QHBoxLayout()
            slider_layout.addWidget(label)
            slider_layout.addWidget(slider)
            slider_widget = QWidget()
            slider_widget.setLayout(slider_layout)
            main_layout.addWidget(slider_widget)
            self.slider_widgets.append(slider_widget)
            if dim not in viewer.dims.order[-viewer.dims.ndisplay:]:
                slider_widget.setVisible(False)
            self.sliders.append(slider)
        grid_layout = QGridLayout()
        self.projections = []
        for dim_pair in itertools.combinations(range(image_layer.ndim), 2):
            if channels_dim in dim_pair:
                continue
            if dim_pair[-2] == viewer.dims.order[-1] or dim_pair[-1] == viewer.dims.order[-2]:
                dim_pair = tuple(reversed(dim_pair))
                flip = True
            else:
                flip = False
            slice_indices = layer_slice_indices(image_layer)
            slices = [max(min(int(slice_indices[dim]), image_layer.data.shape[dim]-1), 0) if dim in layer_dims_not_displayed(image_layer) else 0 for dim in range(viewer.dims.ndim)]
            projection = DataProjectionWidget(viewer, image_layer, mask_layer, dim_pair, slices=slices, flip_image=flip, crosshair_color=self.overlay_color)
            grid_layout.setRowStretch(len(self.projections)//3, 1)
            grid_layout.setColumnStretch(len(self.projections)%3, 1)
            grid_layout.addWidget(projection, len(self.projections)//3, len(self.projections)%3, QtCore.Qt.AlignmentFlag.AlignCenter)
            self.projections.append(projection)
        self.grid_widget = QWidget()
        self.grid_widget.setLayout(grid_layout)
        self.grid_widget.setSizePolicy(self.grid_widget.sizePolicy().horizontalPolicy(), QSizePolicy.Expanding)
        main_layout.addWidget(self.grid_widget)
        extra_settings_layout = QHBoxLayout()
        self.dockable_checkbox = QCheckBox("dockable")
        dockable = self.settings.value("projectionsWidgetDockable", 'false')
        self.dockable_checkbox.setChecked(dockable == "true")
        self.dockable_checkbox.clicked.connect(self.set_dockable)
        extra_settings_layout.addWidget(self.dockable_checkbox)
        self.change_color_button = QPushButton()
        self.change_color_button.setFlat(True)
        self.change_color_button.setStyleSheet("background-color: %s;" % colors.to_hex(list(map(lambda v: v/255, self.overlay_color))))
        self.change_color_button.clicked.connect(self.show_overlay_colorpicker)
        self.change_color_button.setFixedWidth(20)
        self.change_color_button.setFixedHeight(20)
        self.setToolTip("Change overlay color")

        extra_settings_layout.addWidget(self.change_color_button)
        main_layout.addLayout(extra_settings_layout)
        self.setLayout(main_layout)
        self.setSizePolicy(self.grid_widget.sizePolicy().horizontalPolicy(), QSizePolicy.Expanding)
        viewer.dims.events.current_step.connect(self.on_step_change)
        viewer.dims.events.order.connect(self.on_order_change)
        mask_layer.mouse_drag_callbacks.append(self.update_layer)
        mask_layer.mouse_move_callbacks.append(self.show_crosshair)

    def show_crosshair(self, layer, event):
        if self.viewer.dims.ndisplay == 3:
            return

        float_coordinates = layer.world_to_data(event.position)
        coordinates = np.round(np.asarray(float_coordinates)).astype(int)
        for p in self.projections:
            p.update_overlay(layer, coordinates)
        for i, slider in enumerate(self.sliders):
            if i in self.viewer.dims.order[-self.viewer.dims.ndisplay:] and layer.data.shape[i] > coordinates[i] >= 0:
                slider.setValue(coordinates[i])

    def on_step_change(self, event=None, layers=None):
        for projection in self.projections:
            projection.setSlices(*itertools.chain(*((dim, int(event.source.current_step[dim]-self.offset[dim])) for dim in event.source.not_displayed)))

    def update_layer(self, layer, event):
        self.show_crosshair(layer, event)
        for projection in self.projections:
            projection.update()
        yield
        while event.type == "mouse_move":
            self.show_crosshair(layer, event)
            for projection in self.projections:
                projection.update()
            yield
        for projection in self.projections:
            projection.update()

    def update_slider_ranges(self):
        for dim in range(len(self.sliders)):
            self.sliders[dim].setMaximum(self.image_layer.data.shape[dim])

    def slider_callback(self, dim):
        def on_slider_change(val):
            for projection in self.projections:
                projection.setSlices(dim, val)
        return on_slider_change

    def on_order_change(self, _):
        for dim, slider in enumerate(self.slider_widgets):
            slider.setVisible(dim in self.viewer.dims.order[-self.viewer.dims.ndisplay:])
        for p in self.projections:
            if p.displayed_axes[-2] == self.viewer.dims.order[-1] or p.displayed_axes[-1] == self.viewer.dims.order[-2]:
                p.flip_image = not p.flip_image
                p.displayed_axes = type(p.displayed_axes)(reversed(p.displayed_axes))

    def sizeHint(self) -> QtCore.QSize:
        return QtCore.QSize(640, 480)

    def show_overlay_colorpicker(self):
        color = QColorDialog.getColor(QColor(*self.overlay_color), options=QColorDialog.ShowAlphaChannel)
        if color.isValid():
            self.overlay_color = color.red(), color.green(), color.blue(), color.alpha()
            self.settings.setValue("projection_overlay_color", self.overlay_color)
            self.change_color_button.setStyleSheet("background-color: %s;" % colors.to_hex(list(map(lambda v: v/255, self.overlay_color))))
            for projection in self.projections:
                projection.crosshair_color = self.overlay_color

    def set_dockable(self, state):
        self.parent().setAllowedAreas(Qt.AllDockWidgetAreas if state else Qt.NoDockWidgetArea)

    def showEvent(self, QShowEvent):
        if self.shown:
            return
        self.shown = True
        dock_widget = self.parent()
        dock_widget.setFeatures(dock_widget.features() & ~QDockWidget.DockWidgetClosable)
        geometry = self.settings.value("projectionsWidgetGeometry", None)
        is_floating = self.settings.value("projectionsWidgetFloating", None)
        if is_floating is not None:
            dock_widget.setFloating(is_floating == 'true')
        if geometry is not None:
            dock_widget.setGeometry(geometry)
        self.set_dockable(self.dockable_checkbox.isChecked())

    def hideEvent(self, QHideEvent):
        self.settings.setValue("projectionsWidgetGeometry", self.parent().geometry())
        self.settings.setValue("projectionsWidgetFloating", self.parent().isFloating())
        self.settings.setValue("projectionsWidgetDockable", self.dockable_checkbox.isChecked())
        super().hideEvent(QHideEvent)



# cython: language_level=3, language = c++
cimport cython
from cython.parallel cimport prange
import numpy as np
cimport numpy as np
np.import_array()
from libc.math cimport fabs, sqrt, cbrt
from ._essentials import dt, magnitude, innerProduct, ReconstructionMethods
from scipy.sparse.linalg import svds
from scipy.sparse import csc_matrix
from scipy.optimize import minimize
from scipy.integrate import solve_bvp
from .rk import rk4
from .interpHelper import *
import time
import matplotlib.pyplot as plt
Q_interp = InterpHelper()
Qdot_interp = InterpHelper()
theta_interp = InterpHelper()
thetadot_interp = InterpHelper()
thetaddot_interp = InterpHelper()


# function that is called from the main executor
def reconstruct(np.ndarray[np.double_t, ndim=2] q_mean, np.ndarray[np.double_t, ndim=1] guessRayLengths, settings, signal):
    if settings.reconMethod is ReconstructionMethods.NEWTON:
        return reconstruct_newton(q_mean, guessRayLengths, settings.debug, signal)
    if settings.reconMethod is ReconstructionMethods.GRADIENT_DESCENT:
        return gDescent(q_mean,guessRayLengths, settings)
    if settings.reconMethod is ReconstructionMethods.CG:
        return reconstruct_cgradient(q_mean, guessRayLengths, settings, signal)
    if settings.reconMethod is ReconstructionMethods.SKIP:
        return reconstruct_iterate_implicit(q_mean,guessRayLengths)
    if settings.reconMethod is ReconstructionMethods.JOZSI_GRADIENT:
        return reconstruct_gradient(q_mean, guessRayLengths, 100, settings.debug, signal)
    raise NotImplementedError("Method %s not implemented" % settings.reconMethod)


# cost function for gradient descent, x: |r|, qraylengths: |q|, dirs are the unit direction vectors for q and r.
def cost(x, qraylengths, dirs):
    rguess = dirs*x.reshape(x.shape[0],1)
    rdot = dt(rguess,1)
    rdot_abs = magnitude(rdot)
    srv = np.sqrt(rdot_abs)
    return (qraylengths-x*srv)**2

# derivative of the cost function with respect to x (|r|)
def dcost(x, qraylengths, dirs):
    rguess = dirs*x.reshape(x.shape[0],1) # r
    q = dirs*qraylengths.reshape(qraylengths.shape[0],1) # q

    rdot = dt(rguess,1)
    rdotdot = dt(rguess,2)
    rdot_abs = magnitude(rdot)
    qdot = dt(q,1)
    srv = np.sqrt(rdot_abs)

    nominator = innerProduct(q,qdot)/qraylengths - ( (innerProduct(rguess,rdot)/x)*srv + (x/(2*srv))*(innerProduct(rdot,rdotdot)/rdot_abs) )
    denominator = innerProduct(rguess,rdot)

    gradients = x*(nominator/denominator)
    gradients *= 2*(qraylengths-x*srv)

    return gradients

# gradient descent algorithm
def gDescent(q, guessRayLengths, settings):
    qraylengths = np.sqrt(innerProduct(q,q))
    dirs = q/qraylengths.reshape(qraylengths.shape[0],1)
    alpha = settings.alpha
    iters = settings.gradientIterations
    maxgrad = 1.0
    costs = []
    for i in range(iters):
        # calculate cost
        cost_sum = np.sum(cost(guessRayLengths,qraylengths,dirs))
        print("cost = "+str(cost_sum))
        costs.append(cost_sum)

        # calculate the derivative of the cost wrt |r|
        gradients = dcost(guessRayLengths, qraylengths, dirs)

        # make sure that no gradient is too large, rescale them so that the biggest gradient is of size maxgrad at most.
        curr_max = np.max(np.abs(gradients))
        if (curr_max>maxgrad):
            gradients /= curr_max
            gradients *= maxgrad
        
        # gradient descent update rule
        guessRayLengths = guessRayLengths-alpha*gradients
        #guessNext = np.roll(guessRayLengths,-1,axis=0)
        #guessPrev = np.roll(guessRayLengths,1,axis=0)
        #guessRayLengths = 0.1*(8*guessRayLengths+guessNext+guessPrev)

    return guessRayLengths, costs


# gradient descent reconstruction of a q-space curve with an initial guess: will return the length of each ray
def reconstruct_gradient(q_mean, guessRayLengths, multiplier, debug, plotSignal):
    qraylengths = np.sqrt(q_mean[:,0]*q_mean[:,0] + q_mean[:,1]*q_mean[:,1])
    qraylengths[qraylengths<1e-99] = 1e-99
    dirs = q_mean.copy()

    dirs[:,0] /= qraylengths # direction of the mean contour points
    dirs[:,1] /= qraylengths

    startRepr = dirs.copy()
    #if np.abs(q-0.5) < 0.01:
    startRepr[:,0]*=qraylengths
    startRepr[:,1]*=qraylengths
    
    nIter = 10*multiplier
    costs = []
    for i in range(nIter):
        rguess = dirs*guessRayLengths.reshape(guessRayLengths.shape[0],1)
        rdot = dt(rguess,1)
        rdot_abs = magnitude(rdot)
        srv = np.sqrt(rdot_abs)

        energy = qraylengths-guessRayLengths*srv
        energy *= energy
        energy = np.sum(energy, axis=0)
        costs.append(energy)

        rayDiff = qraylengths-(guessRayLengths*srv)

        maxDiff = np.max(np.abs(rayDiff))

        rayDiff *= float(1/maxDiff)

        rayLengths = guessRayLengths + 0.1*rayDiff

        rayNext = np.roll(rayLengths, -1, axis=0)
        rayPrev = np.roll(rayLengths, 1, axis=0)

        #guessRayLengths = rayLengths

        guessRayLengths = 0.1*(8*rayLengths + rayPrev + rayNext)
        
        # display the evolution
        if debug is True and i%10==0:
            cList = []
            restmp = dirs.copy()
            restmp[:,0] *=guessRayLengths*srv
            restmp[:,1] *=guessRayLengths*srv
            cList.append(q_mean[0:q_mean.shape[0]:15,:])
            cList.append(restmp[0:restmp.shape[0]:15,:])
            plotSignal.emit(cList)
            time.sleep(0.1)
        
    return rayLengths, costs

# Newton-Raphson method
def reconstruct_newton(q_mean, guessRayLengths, debug, plotSignal):

    # to avoid a too large condition number
    # guessRayLengths

    bandlength = guessRayLengths.shape[0]-1
    x = guessRayLengths.copy()
    x_prev = guessRayLengths.copy()
    A = np.ndarray((guessRayLengths.shape[0], guessRayLengths.shape[0]), dtype=float)
    b = np.array(guessRayLengths.shape[0], dtype=float)
    upperband = np.array(bandlength, dtype=float)
    lowerband = np.array(bandlength, dtype=float)

    qraylengths = np.sqrt(q_mean[:,0]*q_mean[:,0] + q_mean[:,1]*q_mean[:,1])
    qraylengths[qraylengths<1e-99] = 1e-99
    dirs = q_mean.copy()

    dirs[:,0] /= qraylengths # unit direction of the mean contour points
    dirs[:,1] /= qraylengths

    lrate = 0.01#/guessRayLengths.shape[0]
    nIter = 10*10
    stopCriterion = False
    xmaxglob = 0
    for i in range(nIter):
        if stopCriterion is True:
            break
        dirNext = np.roll(dirs, -1, axis=0)
        dirPrev = np.roll(dirs, 1, axis=0)
        dirNextUnit = np.roll(dirs, -1, axis=0)
        dirPrevUnit = np.roll(dirs, 1, axis=0)
        guessNext = np.roll(guessRayLengths, -1)
        guessPrev = np.roll(guessRayLengths, 1)
        dirNext[:,0]*=guessNext
        dirNext[:,1]*=guessNext
        dirPrev[:,0]*=guessPrev
        dirPrev[:,1]*=guessPrev
        velo = (dirNext-dirPrev)*0.5
        norm_velo = np.sqrt(velo[:,0]*velo[:,0]+velo[:,1]*velo[:,1])
        e = np.ndarray(velo.shape)
        e[:,0] = velo[:,0]/norm_velo
        e[:,1] = velo[:,1]/norm_velo
        velo2 = velo[:,0]*velo[:,0]+velo[:,1]*velo[:,1]
        srv = np.sqrt(np.sqrt(velo2))

        energy = qraylengths-guessRayLengths*np.sqrt(norm_velo)
        energy *= energy
        energy = np.sqrt(np.sum(energy, axis=0))
 
        dirnext_velo_prod = dirNextUnit[:,0]*e[:,0]+dirNextUnit[:,1]*e[:,1]
        dirprev_velo_prod = dirPrevUnit[:,0]*e[:,0]+dirPrevUnit[:,1]*e[:,1]

        upperband = 0.25 * (guessRayLengths[:bandlength-1]/(norm_velo[:bandlength-1])) * dirnext_velo_prod[bandlength-1]
        lowerband = -0.25 * (guessRayLengths[:bandlength-1]/(norm_velo[:bandlength-1])) * dirprev_velo_prod[bandlength-1]
        #upperband = 0.25 * guessRayLengths[:bandlength-1] * (dirnext_velo_prod[:bandlength-1]/velo2[:bandlength-1])
        #lowerband = -0.25 * guessRayLengths[:bandlength-1] * (dirprev_velo_prod[:bandlength-1]/velo2[:bandlength-1])
        A = np.ndarray((guessRayLengths.shape[0], guessRayLengths.shape[0]))
        
        A[:,:] = 0.0
        '''
        print('-----------------------')
        print("min abs upperband: "+str(np.min(np.abs(upperband))))
        print("min abs lowerband: "+str(np.min(np.abs(lowerband))))
        print("max abs upperband: "+str(np.max(np.abs(upperband))))
        print("max abs lowerband: "+str(np.max(np.abs(lowerband))))
        print('-----------------------')
        '''
        np.fill_diagonal(A, 1.0)
        np.fill_diagonal(A[:,1:], upperband)
        np.fill_diagonal(A[1:,:], lowerband)
        A[0,bandlength] = -0.25 * (guessRayLengths[1]/(norm_velo[1])) * dirprev_velo_prod[1]
        A[bandlength,0] = 0.25 * (guessRayLengths[bandlength]/(norm_velo[bandlength])) * dirnext_velo_prod[bandlength]
        b = ( qraylengths / np.sqrt(norm_velo) ) + 0.5*guessRayLengths
        if i==0:
            condnum = np.linalg.cond(A)
            print('-----------------------')
            print("condition number of matrix: "+str(condnum))
            print("min abs upperband: "+str(np.min(np.abs(upperband)))+", min lower "+str(np.min(np.abs(lowerband)))+", max upper "+str(np.max(np.abs(upperband)))+", max lower "+str(np.max(np.abs(lowerband))))
            print("min denominator in A: "+str(np.min(norm_velo))+", max denominator in A: "+str(np.max(norm_velo)))
            print('-----------------------')
        A = csc_matrix(A)

        
        u, s, v = svds(A)
        sm = np.zeros((s.shape[0], s.shape[0]))
        np.fill_diagonal(sm, s)
        x = np.matrix.transpose(v) @ np.linalg.inv(sm) @ np.matrix.transpose(u) @ b
        
        #x = np.linalg.solve(A,b)
        
        # smoothing
        #xn = np.roll(x, -1)
        #x2n = np.roll(x,-2)
        #xp = np.roll(x, 1)
        #x2p = np.roll(x,2)
        #x = (2*xn+x2n+2*xp+x2p+4*x)/10
        #----------------
        
        if np.max(np.abs(x)) > xmaxglob:
            xmaxglob = np.max(np.abs(x))


        x = (1-lrate)*x_prev + lrate*x

        guessRayLengths = x.copy()

        if debug is True and i%5==0:
            cList = []
            restmp = dirs.copy()
            dirNext = np.roll(dirs, -1, axis=0)
            dirPrev = np.roll(dirs, 1, axis=0)
            guessNext = np.roll(guessRayLengths, -1)
            guessPrev = np.roll(guessRayLengths, 1)
            dirNext[:,0]*=guessNext
            dirNext[:,1]*=guessNext
            dirPrev[:,0]*=guessPrev
            dirPrev[:,1]*=guessPrev
            velo = (dirNext-dirPrev)*0.5
            norm_velo = np.sqrt(velo[:,0]*velo[:,0]+velo[:,1]*velo[:,1])
            srv = np.sqrt(norm_velo)
            restmp[:,0] *=guessRayLengths#*srv
            restmp[:,1] *=guessRayLengths#*srv
            cList.append(q_mean[0:q_mean.shape[0]:15,:])
            cList.append(restmp[0:restmp.shape[0]:15,:])
            plotSignal.emit(cList)
            time.sleep(0.1)


        xdiff = x-x_prev
        absxdiffmean = np.mean(np.sqrt(xdiff*xdiff))
        if absxdiffmean < 0.001:
            stopCriterion = True
        x_prev = x.copy()
    print("max x during opt:"+str(xmaxglob))
    print("x max after:"+str(np.max(np.abs(guessRayLengths))))
    return guessRayLengths

# cost function of the reconstruction: used at the conjugate gradient method as this is a sum
def energyfunc(x, qraylengths, dirs, lambd):

    rguess = dirs*x.reshape(x.shape[0],1)
    velo = magnitude(dt(rguess,1))
    curvature = magnitude(dt(rguess,2))
    kappanom = np.abs(dt(rguess[:,0],1)*dt(rguess[:,1],2)-dt(rguess[:,1],1)*dt(rguess[:,0],2))
    kappadenom = (dt(rguess[:,0],1)**2 + dt(rguess[:,1],1)**2)**(3/2)
    kappa = kappanom/kappadenom
    energy = qraylengths-x*np.sqrt(velo)
    energy *= energy

    energy = lambd*energy+(1-lambd)*kappa

    return np.sum(energy, axis=0)


# conjugate gradient reconstruction
def reconstruct_cgradient(q_mean, guessRayLengths, settings, plotSignal=None):
    #guessRayLengths = reconstruct_gradient(q_mean, guessRayLengths.copy(), debug, plotSignal)
    
    qraylengths = np.sqrt(q_mean[:,0]*q_mean[:,0] + q_mean[:,1]*q_mean[:,1])
    qraylengths[qraylengths<1e-99] = 1e-99

    dirs = q_mean.copy()

    dirs /= qraylengths.reshape(qraylengths.shape[0], 1) # direction of the mean contour points

    lambd = settings.lambd

    opres = minimize(energyfunc, guessRayLengths, args=(qraylengths, dirs, lambd), method='CG',options={'maxiter': 1000, 'eps': 1.4901161193847656e-8})
    print("final cost: "+str(opres.fun))
    x = opres.x
    #x = optimize_gradient(guessRayLengths,qraylengths,dirs)

    return x, []

# for R-K reconstruction
def getPotentialZeroPoints(q, guessRayLengths):
    q_abs = magnitude(q)
    q_abs_dot = dt(q_abs,1)
    q_x = q[:,0]
    q_y = q[:,1]
    q_x_dot = dt(q_x, 1)
    q_y_dot = dt(q_y, 1)
    q_x_ddot = dt(q_x, 2)
    q_y_ddot = dt(q_y, 2)
    theta = np.arctan2(q_y,q_x)
    theta_dot = dt(theta, 1)#(q_y_dot*q_x - q_y*q_x_dot)/(q_x*q_x+q_y*q_y)
    theta_ddot = dt(theta, 2)#(q_x*q_y_ddot - q_y*q_x_ddot)/(q_x*q_x+q_y*q_y) -2*((q_x*q_y_dot-q_y*q_x_dot)*(q_y*q_y_dot+q_x*q_x_dot))/(q_x*q_x+q_y*q_y)**2

    diffs = 2*q_abs_dot*theta_dot - q_abs*theta_ddot
    pointnum = q_abs.shape[0]
    candidate_num = diffs[diffs<1].shape[0]
    print(diffs)

    idxes = np.argwhere(theta_dot<-2)
    theta_dot[theta_dot<-0.5] = 0
    
    indices = np.argsort(diffs)
    f = plt.figure()
    plt.plot(q[:,1], -1*q[:,0])
    plt.scatter(q[idxes,1], -1*q[idxes,0], label="theta_dot<-2", color="r")
    plt.legend()
    f.savefig("candidates.png", dpi=300)

    r_abs_dot = dt(guessRayLengths,1)
    r_abs_ddot = dt(guessRayLengths,2)
    t = np.linspace(0,len(guessRayLengths)-1,len(guessRayLengths))
    f = plt.figure()
    plt.subplot(1,2,1)
    plt.title("r_abs_dot")
    plt.plot(t,r_abs_dot)
    plt.subplot(1,2,2)
    plt.title("theta_dot^2")
    plt.plot(t,theta_dot**2)
    f.savefig("guessderivs.png", dpi=300)

    zeroplaces = np.argwhere(np.abs(r_abs_dot)<0.03)
    print(zeroplaces.shape)
    f = plt.figure()
    plt.plot(q[:,1], -1*q[:,0])
    plt.scatter(q[zeroplaces,1], -1*q[zeroplaces,0], label="r_abs_dot~=0", color="r")
    plt.legend()
    f.savefig("zeroplaces.png", dpi=300)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cpdef reconstruct_iterate_implicit(np.ndarray[np.double_t, ndim=2] q_mean, np.ndarray[np.double_t, ndim=1] guessRayLengths):
    cdef np.ndarray[np.double_t, ndim=1] r_abs = guessRayLengths.copy()
    cdef np.ndarray[np.double_t, ndim=2] u = q_mean/magnitude(q_mean).reshape(q_mean.shape[0],1)
    cdef np.ndarray[np.double_t, ndim=2] r = np.empty_like(q_mean)
    cdef np.ndarray[np.double_t, ndim=2] r_dot
    cdef np.ndarray[np.double_t, ndim=1] r_dot_abs
    cdef np.ndarray[np.double_t, ndim=2] e
    cdef np.ndarray[np.double_t, ndim=1] q_abs = magnitude(q_mean)
    cdef np.ndarray[np.double_t, ndim=1] diffs = np.empty_like(q_abs)
    cdef np.ndarray[np.double_t, ndim=1] ue_inner = np.empty_like(q_abs)
    cdef np.ndarray[np.double_t, ndim=1] q_abs_sq_ue = np.empty_like(q_abs)  # = q_abs**2 * ue_inner
    cdef np.ndarray[np.double_t, ndim=1] q_abs_sq_ue_cumsum  # cumulative sum of q_abs_sq_ue
    cdef double first, second, third, q_abs_sq_ue_sum
    cdef int i, t, j
    cdef double init_len = 0
    cdef double loop_len = 0
    cdef double r_abs_, u_, r_dot_abs_, e1_, e2_, q_abs_, diffs_
    for i in range(100):
        q_abs_sq_ue_sum = 0
        for j in prange(r_abs.shape[0], nogil=True):
            r[j, 0] = u[j, 0] * r_abs[j]
            r[j, 1] = u[j, 1] * r_abs[j]
        r_dot = dt(r, 1)
        for j in prange(r_abs.shape[0], nogil=True):
            r_dot_abs_ = sqrt(r_dot[j, 0]*r_dot[j, 0] + r_dot[j, 1]*r_dot[j, 1])
            e1_ = r_dot[j, 0]/r_dot_abs_
            e2_ = r_dot[j, 1]/r_dot_abs_
            ue_inner[j] = fabs(e1_*u[j, 0]+e2_*u[j, 1])
            q_abs_sq_ue[j] = q_abs[j]*q_abs[j]*ue_inner[j]
            q_abs_sq_ue_sum += q_abs_sq_ue[j]
        q_abs_sq_ue_cumsum = np.cumsum(q_abs_sq_ue)
        for t in prange(q_abs.shape[0], nogil=True):
        # for t in range(q_abs.shape[0]):
            first = 3.0*q_abs_sq_ue_cumsum[t]
            # print("msg1")
            second = (( q_abs_sq_ue[0] / fabs(u[0, 0]*r_dot[0, 0]+u[0, 1]*r_dot[0, 1]))**(3/2))
            # print("msg2")
            third = -3.0*q_abs_sq_ue_sum
            # print("msg3")
            diffs[t] = (fabs(cbrt(first+second+third)) -r_abs[t])
            # print("msg4")
        maxdiff = np.max(diffs)
        if maxdiff>1:
            diffnorm = diffs/maxdiff
        else:
            diffnorm = diffs
        r_abs += diffnorm*0.01

    return r_abs, []


def rk_f(t, y, Q,Q_dot,theta,theta_dot):
    #print("t = "+str(t))
    # (np.sign(theta_dot[t])*dt(theta,2)[t])/np.abs(theta_dot[t])
    theta_ddot = dt(theta,2)

    return 3*np.abs(theta_dot[t])+np.tan(y)*( theta_ddot[t]/theta_dot[t] - (2*Q_dot[t]/Q[t]))

def reconstruct_rk(q_mean, guessRayLengths):
    dirs = q_mean/magnitude(q_mean).reshape(q_mean.shape[0],1)
    Q = magnitude(q_mean)
    Q_dot = dt(Q,1)
    theta = np.arctan2(q_mean[:,1],q_mean[:,0])
    theta_dot = dt(theta,1)
    theta_ddot = dt(theta,2)
    r_guess = dirs*guessRayLengths.reshape(guessRayLengths.shape[0],1)
    r_guess_dot = dt(r_guess,1)
    velo_guess = magnitude(r_guess_dot)


    t = np.linspace(0,Q.shape[0], num=Q.shape[0])
    eqpart = np.sign(theta_dot)*theta_ddot/np.abs(theta_dot)

    '''plt.close('all')
    plt.subplot(1,3,1)
    plt.plot(t,theta, label="theta")
    plt.subplot(1,3,2)
    plt.plot(t,theta_dot, label="theta_dot")
    plt.subplot(1,3,3)
    plt.plot(t,theta_ddot, label="theta_ddot")
    plt.legend()
    plt.show()'''

    v_guess = np.arccos(np.sign(theta_dot) * innerProduct(r_guess,r_guess_dot)/(guessRayLengths*velo_guess))

    '''plt.close('all')
    plt.plot(t,v_guess, label="v guess")
    plt.legend()
    plt.show()'''

    t0 = 100
    v0 = v_guess[t0]
    print("v0 = "+str(v0))
    t, my_guess_v, tans = rk4(rk_f,t0,v0,Q,Q_dot,theta,theta_dot,2,399)

    

    #------------- ver 2
    #my_guess_y = my_guess_v
    #sinv = np.sqrt(1/(1+1/(my_guess_y)**2))

    #--------------------


    r_lengths = np.cbrt(np.sin(my_guess_v)*((Q[t0:len(my_guess_v)+t0]**2)/np.abs(theta_dot[t0:len(my_guess_v)+t0])))
    #print(my_guess_v)
    curve_guess = dirs[t0:2*len(my_guess_v)+t0:2]*r_lengths.reshape(r_lengths.shape[0],1)

    f = plt.figure()
    plt.subplot(1,2,1)
    plt.plot(curve_guess[:,1],-1*curve_guess[:,0], label="rk")
    plt.plot(r_guess[t0:len(my_guess_v)+t0,1],-1*r_guess[t0:len(my_guess_v)+t0,0], label="guess")
    plt.legend()

    t = np.linspace(0,len(my_guess_v),num=len(my_guess_v))
    plt.subplot(1,2,2)
    plt.plot(t,my_guess_v)
    f.savefig("rk.png", dpi=300)
    plt.close(f)
    
    


def makePeriodic(arr):
    if len(arr.shape)>1:
        ret_arr = np.ndarray((arr.shape[0]+1,2))
    else:
        ret_arr = np.ndarray((arr.shape[0]+1,))
    
    ret_arr[:arr.shape[0]] = arr
    ret_arr[-1] = arr[0]

    return ret_arr

def dydt(t,y):
    return 3*np.abs(thetadot_interp.interpolator(t))+np.tan(y)*( thetaddot_interp.interpolator(t)/thetadot_interp.interpolator(t)- (2*Qdot_interp.interpolator(t)/Q_interp.interpolator(t)) )

def bc(ya,yb):
    return ya-yb

def reconstruct_bvp(q_mean, guessRayLengths):
    dirs = q_mean/magnitude(q_mean).reshape(q_mean.shape[0],1)
    Q = magnitude(q_mean)
    Q_dot = dt(Q,1)
    theta = np.arctan2(q_mean[:,1],q_mean[:,0])
    theta_dot = dt(theta,1)
    theta_ddot = dt(theta,2)
    r_guess = dirs*guessRayLengths.reshape(guessRayLengths.shape[0],1)
    r_guess_dot = dt(r_guess,1)
    velo_guess = magnitude(r_guess_dot)
    v_guess = np.arccos(np.sign(theta_dot) * innerProduct(r_guess,r_guess_dot)/(guessRayLengths*velo_guess))

    Q = makePeriodic(Q)
    Q_dot = makePeriodic(Q_dot)
    theta = makePeriodic(theta)
    theta_dot = makePeriodic(theta_dot)
    theta_ddot = makePeriodic(theta_ddot)
    r_guess = makePeriodic(r_guess)
    r_guess_dot = makePeriodic(r_guess_dot)
    velo_guess = makePeriodic(velo_guess)
    v_guess = makePeriodic(v_guess)
    v_guess = v_guess.reshape(1,v_guess.shape[0])

    t = np.linspace(0,Q.shape[0], num=Q.shape[0])
    Q_interp.setInterpolator(t,Q)
    Qdot_interp.setInterpolator(t,Q_dot)
    theta_interp.setInterpolator(t,theta)
    thetadot_interp.setInterpolator(t,theta_dot)
    thetaddot_interp.setInterpolator(t,theta_ddot)

    t0 = 100

    res = solve_bvp(dydt,bc,t,v_guess, max_nodes=100000)
    print("status of result: "+str(res.status))

    tres = res.x
    vres = res.y.reshape(res.y.shape[1],)
    R = np.cbrt( ((Q_interp.interpolator(tres)**2)/np.abs(Qdot_interp.interpolator(tres))) * np.sin(vres))
    f = plt.figure()
    plt.plot(tres,R)
    f.savefig("rtest.png", dpi=300)
    plt.close(f)
    f = plt.figure()
    plt.plot(tres,vres)
    f.savefig("vtest.png", dpi=300)
    plt.close(f)


